import pandas as pd
from pandas.testing import assert_frame_equal
import numpy as np

import pytest
from pytest import approx
from pytest import mark

import okama as ok

from .conftest import data_folder


@mark.asset_list
def test_asset_list_init_failing():
    with pytest.raises(ValueError, match=r"Assets must be a list."):
        ok.AssetList(assets=("RUB.FX", "MCFTR.INDX"))


@mark.asset_list
@mark.usefixtures("_init_asset_list")
class TestAssetList:
    def test_repr(self):
        value = pd.Series(dict(
            assets="[pf1.PF, RUB.FX, MCFTR.INDX]",
            currency="USD",
            first_date="2019-02",
            last_date="2020-01",
            period_length="1 years, 0 months",
            inflation="USD.INFL"
        ))
        assert repr(self.asset_list_with_portfolio) == repr(value)

    def test_len(self):
        assert self.asset_list.__len__() == 2

    def test_tickers(self):
        assert self.asset_list_with_portfolio.tickers == ['pf1', 'RUB', 'MCFTR']

    def test_ror(self):
        asset_list_sample = pd.read_pickle(data_folder / "asset_list.pkl")
        asset_list_lt_sample = pd.read_pickle(data_folder / "asset_list_lt.pkl")
        currencies_sample = pd.read_pickle(data_folder / "currencies.pkl")
        real_estate_sample = pd.read_pickle(data_folder / "real_estate.pkl")
        spy_sample = pd.read_pickle(data_folder / "spy.pkl")
        assert_frame_equal(self.asset_list.assets_ror, asset_list_sample)
        assert_frame_equal(self.asset_list_lt.assets_ror, asset_list_lt_sample)
        assert_frame_equal(self.currencies.assets_ror, currencies_sample)
        assert_frame_equal(self.real_estate.assets_ror, real_estate_sample)
        # SPY adj_close and ror is changing over time
        assert_frame_equal(self.spy.assets_ror, spy_sample, rtol=1e-2)

    def test_currencies(self):
        assert self.currencies.pl.years == 1
        assert self.currencies.first_date == pd.to_datetime("2019-01")
        assert self.currencies.currencies == {
            "RUBUSD.FX": "USD",
            "EURUSD.FX": "USD",
            "CNYUSD.FX": "USD",
            "asset list": "USD",
        }
        assert self.currencies.names == {
            "RUBUSD.FX": "RUBUSD",
            "EURUSD.FX": "EURUSD",
            "CNYUSD.FX": "CNYUSD",
        }
        assert self.currencies.describe().iloc[1, -1] == approx(0.02485, rel=1e-2)

    def test_names(self):
        assert list(self.spy.names.values()) == ["SPDR S&P 500 ETF Trust"]

    @mark.smoke
    def test_make_asset_list(self):
        assert self.asset_list.last_date == pd.to_datetime("2020-01")
        assert list(self.asset_list.assets_ror) == ["RUB.FX", "MCFTR.INDX"]

    def test_calculate_wealth_indexes(self):
        assert self.asset_list.wealth_indexes.sum(axis=1)[-1] == approx(
            3339.677963676333, rel=1e-2
        )  # last month indexes sum

    def test_risk(self):
        assert self.asset_list.risk_monthly["RUB.FX"] == approx(0.0258, rel=1e-2)
        assert self.asset_list.risk_monthly["MCFTR.INDX"] == approx(0.0264, rel=1e-2)
        assert self.asset_list.risk_annual["RUB.FX"] == approx(0.0825, rel=1e-2)
        assert self.asset_list.risk_annual["MCFTR.INDX"] == approx(0.1222, rel=1e-2)

    def test_semideviation_monthly(self):
        assert self.asset_list.semideviation_monthly[0] == approx(0.015614, rel=1e-2)
        assert self.asset_list.semideviation_monthly[1] == approx(0, abs=1e-2)

    def test_semideviation_annual(self):
        assert self.asset_list.semideviation_annual[0] == approx(0.05408, rel=1e-2)
        assert self.asset_list.semideviation_annual[1] == approx(0, abs=1e-2)

    def test_get_var_historic(self):
        assert self.asset_list.get_var_historic(time_frame=1, level=5)["RUB.FX"] == approx(0.0411, rel=1e-2)
        assert self.asset_list.get_var_historic(time_frame=5, level=1)["MCFTR.INDX"] == approx(-0.1048, rel=1e-2)
        assert self.asset_list_no_infl.get_var_historic(time_frame=1, level=1)["RUB.FX"] == approx(0.04975, rel=1e-2)
        assert self.asset_list_no_infl.get_var_historic(time_frame=1, level=1)["MCFTR.INDX"] == approx(0.01229, rel=1e-2)

    @mark.test
    def test_get_cvar_historic(self):
        assert self.asset_list.get_cvar_historic(level=5, time_frame=12)[
            "RUB.FX"
        ] == approx(0.1120, rel=1e-2)
        assert self.asset_list.get_cvar_historic(level=5, time_frame=12)[
            "MCFTR.INDX"
        ] == approx(-0.3130, rel=1e-2)

    def test_drawdowns(self):
        assert self.asset_list.drawdowns.min().sum() == approx(-0.082932, rel=1e-2)

    def test_recovery_periods(self):
        assert self.asset_list.recovery_periods['MCFTR.INDX'] == approx(0, rel=1e-2)
        assert np.isnan(self.asset_list.recovery_periods['RUB.FX'])
        assert self.asset_list_lt.recovery_periods['MCFTR.INDX'] == 45
        assert self.asset_list_lt.recovery_periods['RUB.FX'] == 69

    cagr_testdata1 = [
        (1, -0.0463, 0.3131, 0.0242),
        (None, -0.0888, 0.3651, 0.0318),
    ]

    @mark.parametrize(
        "input_data,expected1,expected2,expected3",
        cagr_testdata1,
        ids=["1 year", "full period"],
    )
    def test_get_cagr(self, input_data, expected1, expected2, expected3):
        assert self.asset_list.get_cagr(period=input_data)["RUB.FX"] == approx(
            expected1, rel=1e-2
        )
        assert self.asset_list.get_cagr(period=input_data)["MCFTR.INDX"] == approx(
            expected2, rel=1e-2
        )
        assert self.asset_list.get_cagr(period=input_data)["RUB.INFL"] == approx(
            expected3, rel=1e-2
        )

    cagr_testdata2 = [
        (1, -0.0688, 0.2820),
        (None, -0.1169, 0.3228),
    ]

    @mark.parametrize("input_data,expected1,expected2", cagr_testdata2, ids=["1 year", "full period"],)
    def test_get_cagr_real(self, input_data, expected1, expected2):
        assert self.asset_list.get_cagr(period=input_data, real=True)["RUB.FX"] == approx(expected1, abs=1e-2)
        assert self.asset_list.get_cagr(period=input_data, real=True)["MCFTR.INDX"] == approx(expected2, abs=1e-2)

    def test_get_cagr_value_error(self):
        with pytest.raises(ValueError):
            self.asset_list.get_cagr(period=3, real=True)

    def test_get_cagr_real_no_inflation_exception(self):
        with pytest.raises(ValueError):
            self.asset_list_no_infl.get_cagr(period=1, real=True)

    @pytest.mark.parametrize(
        "real, expected1, expected2", [(False, 0.05822, 0.2393), (True, 0.0204, 0.1951)]
    )
    def test_get_rolling_cagr(self, real, expected1, expected2):
        assert self.asset_list_lt.get_rolling_cagr(window=24, real=real)["RUB.FX"].iloc[
            -1
        ] == approx(expected1, rel=1e-2)
        assert self.asset_list_lt.get_rolling_cagr(window=24, real=real)[
            "MCFTR.INDX"
        ].iloc[-1] == approx(expected2, rel=1e-2)

    get_rolling_cagr_error_data = [
        (0, False, ValueError),  # window should be at least 12 months for CAGR
        (12.5, False, ValueError),  # not an integer
        (10 * 12, False, ValueError),  # window size should be in the history period
        (12, True, ValueError),  # real CAGR is defined when AssetList(inflation=True) only
    ]

    @pytest.mark.parametrize("window, real, exception", get_rolling_cagr_error_data)
    def test_get_rolling_cagr_error(self, window, real, exception):
        with pytest.raises(exception):
            self.asset_list_no_infl.get_rolling_cagr(window=window, real=real)

    cumulative_testdata1 = [
        ("YTD", 0.0182, 0.0118, 0.0040),
        (1, -0.0463, 0.3131, 0.0242),
        (None, -0.0957, 0.4009, 0.0345),
    ]

    @mark.parametrize(
        "input_data,expected1,expected2,expected3",
        cumulative_testdata1,
        ids=["YTD", "1 year", "full period"],
    )
    def test_get_cumulative_return(self, input_data, expected1, expected2, expected3):
        assert self.asset_list.get_cumulative_return(period=input_data)[
            "RUB.FX"
        ] == approx(expected1, rel=1e-2)
        assert self.asset_list.get_cumulative_return(period=input_data)[
            "MCFTR.INDX"
        ] == approx(expected2, rel=1e-2)
        assert self.asset_list.get_cumulative_return(period=input_data)[
            "RUB.INFL"
        ] == approx(expected3, rel=1e-2)

    cumulative_testdata2 = [
        ("YTD", 0.01424, 0.0077),
        (1, -0.06885, 0.2820),
        (None, -0.1260, 0.3541),
    ]

    @mark.parametrize(
        "input_data,expected1,expected2",
        cumulative_testdata2,
        ids=["YTD", "1 year", "full period"],
    )
    def test_get_cumulative_return_real(self, input_data, expected1, expected2):
        assert self.asset_list.get_cumulative_return(period=input_data, real=True)["RUB.FX"] == approx(expected1, abs=1e-2)
        assert self.asset_list.get_cumulative_return(period=input_data, real=True)["MCFTR.INDX"] == approx(expected2, abs=1e-2)

    def test_get_cumulative_return_value_error(self):
        with pytest.raises(ValueError):
            self.asset_list.get_cumulative_return(period=3, real=True)

    def test_get_cumulative_return_real_no_inflation_exception(self):
        with pytest.raises(ValueError):
            self.asset_list_no_infl.get_cumulative_return(period=1, real=True)

    def test_get_rolling_cumulative_return(self):
        assert self.asset_list_lt.get_rolling_cumulative_return(window=12)[
            "RUB.FX"
        ].iloc[-1] == approx(-0.0462, rel=1e-2)
        assert self.asset_list_lt.get_rolling_cumulative_return(window=12)[
            "MCFTR.INDX"
        ].iloc[-1] == approx(0.3130, rel=1e-2)

    def test_mean_return(self):
        assert self.asset_list.mean_return["RUB.FX"] == approx(-0.0854, rel=1e-2)
        assert self.asset_list.mean_return["MCFTR.INDX"] == approx(0.3701, rel=1e-2)
        assert self.asset_list.mean_return["RUB.INFL"] == approx(0.0319, rel=1e-2)

    def test_real_return(self):
        assert self.asset_list.real_mean_return["RUB.FX"] == approx(-0.11366, rel=1e-2)
        assert self.asset_list.real_mean_return["MCFTR.INDX"] == approx(
            0.3276, rel=1e-2
        )

    def test_annual_return_ts(self):
        assert self.asset_list.annual_return_ts.iloc[-1, 0] == approx(0.01829, rel=1e-2)
        assert self.asset_list.annual_return_ts.iloc[-1, 1] == approx(0.01180, rel=1e-2)

    def test_describe(self):
        description = self.asset_list.describe(tickers=False).iloc[:-2, :]  # last 2 rows have fresh lastdate
        description_sample = pd.read_pickle(data_folder / "asset_list_describe.pkl").iloc[:-2, :]
        cols = list(description_sample.columns.values)
        description = description[cols]  # columns order should not be an issue
        assert_frame_equal(description, description_sample)

    def test_dividend_yield(self):
        assert self.spy.assets_dividend_yield.iloc[-1, 0] == approx(0.0125, abs=1e-3)
        assert self.spy_rub.assets_dividend_yield.iloc[-1, 0] == approx(0.01197, abs=1e-3)
        assert self.asset_list.assets_dividend_yield.iloc[:, 0].sum() == 0
        assert self.asset_list_with_portfolio_dividends.assets_dividend_yield.iloc[-1, 0] == approx(0.0394, abs=1e-3)

    def test_dividends_annual(self):
        assert self.spy.dividends_annual.iloc[-2, 0] == approx(
            1.4194999999999998, rel=1e-2
        )
        assert self.asset_list.dividends_annual.iloc[:, 0].sum() == 0

    def test_growing_dividend_years(self):
        assert self.spy.dividend_growing_years.iloc[-1, 0] == 0

    def test_paying_dividend_years(self):
        assert self.spy.dividend_paying_years.iloc[-2, 0] == 2

    def test_get_dividend_mean_growth_rate_valid(self):
        assert self.spy.get_dividend_mean_growth_rate(period=2).iloc[-1] == approx(
            -0.02765, rel=1e-2
        )

    def test_get_dividend_mean_growth_rate_value_err(self):
        with pytest.raises(
            ValueError,
            match="'period' \\(3\\) is beyond historical data range \\(2.0\\)",
        ):
            self.spy.get_dividend_mean_growth_rate(period=3)

    def test_tracking_difference_failing(self):
        with pytest.raises(
            ValueError,
            match="At least 2 symbols should be provided to calculate Tracking Difference.",
        ):
            self.spy.tracking_difference

    def test_tracking_difference(self):
        assert self.asset_list.tracking_difference["MCFTR.INDX"].iloc[-1] == approx(
            0.4967, rel=1e-2
        )

    def test_tracking_difference_annualized(self):
        assert self.asset_list.tracking_difference_annualized.iloc[-1, 0] == approx(
            0.451000, rel=1e-2
        )

    def test_tracking_error(self):
        assert self.asset_list.tracking_error.iloc[-1, 0] == approx(0.19399, rel=1e-2)

    def test_index_corr(self):
        assert self.asset_list.index_corr.iloc[-1, 0] == approx(-0.57634, rel=1e-2)

    def test_index_beta(self):
        assert self.asset_list.index_beta.iloc[-1, 0] == approx(-0.563714, rel=1e-2)

    def test_skewness(self):
        assert self.asset_list.skewness["RUB.FX"].iloc[-1] == approx(0.444343, rel=1e-2)
        assert self.asset_list.skewness["MCFTR.INDX"].iloc[-1] == approx(
            0.24876, rel=1e-2
        )

    def test_rolling_skewness_failing(self):
        with pytest.raises(
            ValueError, match=r"window size is more than data history depth"
        ):
            self.asset_list.skewness_rolling(window=24)

    def test_kurtosis(self):
        assert self.asset_list.kurtosis["RUB.FX"].iloc[-1] == approx(0.89810, rel=1e-2)
        assert self.asset_list.kurtosis["MCFTR.INDX"].iloc[-1] == approx(
            -1.32129, rel=1e-2
        )

    def test_kurtosis_rolling(self):
        assert self.asset_list_lt.kurtosis_rolling(window=24)["RUB.FX"].iloc[
            -1
        ] == approx(1.4425, rel=1e-2)
        assert self.asset_list_lt.kurtosis_rolling(window=24)["MCFTR.INDX"].iloc[
            -1
        ] == approx(-0.11495, rel=1e-2)

    def test_jarque_bera(self):
        assert self.asset_list.jarque_bera["RUB.FX"].iloc[-1] == approx(
            0.84131, rel=1e-2
        )
        assert self.asset_list.jarque_bera["MCFTR.INDX"].iloc[-1] == approx(
            0.60333, rel=1e-2
        )

    def test_get_sharpe_ratio(self):
        sharpe_ratio = self.asset_list.get_sharpe_ratio(rf_return=0.06)
        assert sharpe_ratio.loc['RUB.FX'] == approx(-1.7617, rel=1e-2)
        assert sharpe_ratio.loc['MCFTR.INDX'] == approx(2.53, rel=1e-2)



"""
*okama* is a Python package developed for asset allocation and investments portfolio optimization tasks
according to Modern Portfolio Theory (MPT).
=====================================================================
The package is supplied with **free** «end of day» historical stock markets data and macroeconomic indicators through API.

Main features:

- Investment portfolio constrained Markowitz Mean-Variance Analysis (MVA) and optimization
- Rebalanced portfolio optimization
- Monte Carlo Simulations for financial assets and investment portfolios
- Popular risk metrics: VAR, CVaR, semi-deviation, variance and drawdowns
- Forecasting models according to normal and lognormal distribution
- Testing distribution on historical data
- Dividend yield and other dividend indicators for stocks
- Backtesting and comparing historical performance of broad range of assets and indexes in multiple currencies
- Methods to track the performance of index funds (ETF) and compare them with benchmarks
- Main macroeconomic indicators: inflation, central banks rates
- Matplotlib visualization scripts for the Efficient Frontier, Transition map and assets risk / return performance

"""

from okama.asset import Asset
from okama.asset_list import AssetList
from okama.portfolio import Portfolio
from okama.macro import Inflation, Rate
from okama.frontier.multi_period import EfficientFrontierReb
from okama.frontier.single_period import EfficientFrontier
from okama.api.data_queries import QueryData
from okama.api.search import search
from okama.api.api_methods import API
from okama.api.namespaces import (
    namespaces,
    assets_namespaces,
    macro_namespaces,
    symbols_in_namespace,
)
from okama.common.helpers.helpers import Float, Frame, Rebalance, Date
import okama.settings

__version__ = "1.1.0"



import math
from typing import Union, Callable

import pandas as pd
import numpy as np
import scipy.stats

from okama.settings import _MONTHS_PER_YEAR


def check_rolling_window(
    window: int, ror: Union[pd.Series, pd.DataFrame], window_below_year: bool = False
):
    if not window_below_year and window < _MONTHS_PER_YEAR:
        raise ValueError("window size should be at least 1 year")
    if window > ror.shape[0]:
        raise ValueError("window size is more than data history depth")


class Float:
    """
    Group of methods using float values inputs.
    Some of them can take DataFrame also.
    """

    @staticmethod
    def get_monthly_return_from_annual(
        annual_return: Union[float, pd.Series]
    ) -> Union[float, pd.Series]:
        """
        Gets value of monthly return for a given annual return.
        """
        return (1.0 + annual_return) ** (1 / _MONTHS_PER_YEAR) - 1.0

    @staticmethod
    def annualize_return(
        rate_of_return: Union[float, pd.Series],
        periods_per_year: int = _MONTHS_PER_YEAR,
    ) -> Union[float, pd.Series]:
        """
        Annualizes a return.
        Default annualization is from month to year.
        """
        return (rate_of_return + 1.0) ** periods_per_year - 1.0

    @staticmethod
    def annualize_risk(
        risk: Union[float, pd.Series], mean_return: Union[float, pd.Series]
    ) -> Union[float, pd.Series]:
        """
        Annualizes Risk.
        Annualization from month to year (from standard deviation) is by default. Monthly mean return is also required.
        Works with DataFrame inputs (in math.sqrt is not used).
        """
        return (
            (risk ** 2 + (1 + mean_return) ** 2) ** _MONTHS_PER_YEAR
            - (1 + mean_return) ** (_MONTHS_PER_YEAR * 2)
        ) ** 0.5

    @staticmethod
    def approx_return_risk_adjusted(mean_return: float, std: float) -> float:
        """
        Approximates geometric mean return given mean return and std.
        """
        return (
            np.exp(
                np.log(1.0 + mean_return) - 0.5 * std ** 2 / (1.0 + mean_return) ** 2
            )
            - 1.0
        )

    @staticmethod
    def get_random_weights(n: int, w_shape: int) -> pd.Series:
        """
        Produce N random normalized weights of a given shape.
        """
        # Random weights
        rand_nos = np.random.rand(n, w_shape)
        weights_transposed = rand_nos.transpose() / rand_nos.sum(axis=1)
        weights = weights_transposed.transpose()
        weights_df = pd.DataFrame(weights)
        return weights_df.aggregate(
            np.array, axis=1
        )  # Converts df to DataFrame of np.array

    @staticmethod
    def get_purchasing_power(inflation: float, value: float = 1000.0):
        return value / (1.0 + inflation)


class Frame:
    """
    Group of methods using DataFrame (or Series) input.
    """

    # Rate of return metrics

    @staticmethod
    def weights_sum_is_one(weights: list) -> None:
        if np.around(np.sum(weights), decimals=3) != 1.0:
            raise ValueError("Weights sum is not equal to one.")
        if any(x < 0 for x in weights):
            raise ValueError("Negative weights are not allowed.")

    @staticmethod
    def change_period_to_month(ts: pd.Series) -> pd.Series:
        """
        Change time series period from day to month.
        """
        return ts.resample('M').last()

    @classmethod
    def get_portfolio_return_ts(cls, weights: list, ror: pd.DataFrame) -> pd.Series:
        """
        Returns the mean return time series given portfolio weights and the DataFrame of assets mean returns.
        """
        cls.weights_sum_is_one(weights)
        if isinstance(ror, pd.Series):  # required for a single asset portfolio
            return ror
        return ror @ weights

    @classmethod
    def get_portfolio_mean_return(cls, weights: Union[list, np.array], ror: pd.DataFrame) -> float:
        """
        Computes mean return of a portfolio (monthly).
        """
        # cls.weights_sum_is_one(weights)
        weights = np.asarray(weights)
        if isinstance(ror.mean(), float):  # required for a single asset portfolio
            return ror.mean()
        return weights.T @ ror.mean()

    @staticmethod
    def get_cagr(ror: Union[pd.DataFrame, pd.Series]) -> Union[pd.Series, float]:
        """
        Return Compound Annual Rate of Return (CAGR) for each asset given returns time series DataFrame.
        """
        if ror.shape[0] < 12:
            return pd.Series(
                {x: None for x in ror.columns}
            )  # CAGR is not defined for time periods < 1 year
        return ((ror + 1.0).prod()) ** (_MONTHS_PER_YEAR / ror.shape[0]) - 1.0

    @staticmethod
    def get_cumulative_return(
        ror: Union[pd.DataFrame, pd.Series]
    ) -> Union[pd.Series, float]:
        """
        Return Compound Return for time series of return (one or several).
        """
        return (ror + 1.0).prod() - 1.0

    @staticmethod
    def get_rolling_fn(
        ror: Union[pd.DataFrame, pd.Series],
        window: int,
        fn: Callable,
        window_below_year: bool = False,
    ) -> Union[pd.DataFrame, pd.Series]:
        check_rolling_window(
            window=window, ror=ror, window_below_year=window_below_year
        )
        x = ror.rolling(window).apply(fn)
        return x.dropna()

    @staticmethod
    def get_annual_return_ts_from_monthly(
        ror_monthly: Union[pd.DataFrame, pd.Series]
    ) -> Union[pd.DataFrame, pd.Series]:
        """
        Annual Rate of Returns time series from monthly data.
        """
        ts = ror_monthly.resample("A").apply(lambda x: np.prod(x + 1.0) - 1)
        if isinstance(ts, pd.Series):
            ts.rename(ror_monthly.name, inplace=True)
        return ts

    @staticmethod
    def get_wealth_indexes(
        ror: Union[pd.Series, pd.DataFrame], first_value: float = 1000.0
    ) -> Union[pd.Series, pd.DataFrame]:
        """
        Returns wealth indexes for a list of assets (or for portfolio).
        Works also with pd.Series inputs.
        """
        initial_investments = first_value
        first_date = ror.index[0]
        wealth_index = initial_investments * (ror + 1).cumprod()
        wealth_index.loc[
            first_date
        ] = initial_investments  # replaces NaN with the first period return
        wealth_index.sort_index(ascending=True, inplace=True)
        return wealth_index

    @staticmethod
    def _adjust_rates(rates: pd.Series, symbol: str) -> pd.Series:
        """
        Makes several adjustment to the rates before OKID index is computed:
        - penalty for bank license revocation (except TOP 10 banks)
        - monthly interest capitalization adjustment
        """
        if symbol != "RUS_RUB_TOP10.RATE":
            #  license revocation does not happen among TOP 10 banks
            penalty1 = (
                14 / 365 / 2
            )  # penalty for license revocation (period 1 - frequent revocation)
            penalty2 = (14 / 365) * (
                30 / 440
            )  # penalty for license revocation (period 2 - rare revocation)
            rates["28.09.2018":] = rates["28.09.2018":] * (1 - penalty1)
            rates[:"28.10.2018"] = rates[:"28.10.2018"] * (1 - penalty2)
        # compensate monthly interest capitalization (okid index has it)
        rates = ((rates + 1.0) ** (1 / _MONTHS_PER_YEAR) - 1) * 12.0
        return rates

    @staticmethod
    def get_okid_index(rates: pd.Series, symbol: str) -> pd.Series:
        """
        Computes OKID index from bank rates time series.
        The index is a basket of 12 one year bank deposits (1 month shift).
        symbol is required for rates time series adjustments (for 'RUS_RUB_TOP10.RATE' it's a different adjustment)
        """
        index_total = None
        start_period = pd.Period(rates.index[0], freq="M")
        end_period = pd.Period(rates.index[-1], freq="M")
        rates = Frame._adjust_rates(rates, symbol)
        for month_idx in range(_MONTHS_PER_YEAR):
            rates_yearly = rates.values[month_idx::_MONTHS_PER_YEAR]
            index_part = np.repeat(rates_yearly, _MONTHS_PER_YEAR)[
                : len(rates) - month_idx
            ]
            index_part = (1 + index_part / _MONTHS_PER_YEAR).cumprod()
            index_total = (
                index_part if index_total is None else index_total[1:] + index_part
            )
            start_period += 1
        if index_total is None:
            raise ValueError("`index_total` should not be `None`")
        result = index_total
        result = (np.diff(result) / result[:-1] + 1.0).cumprod() - 1.0
        result = [0] + result
        result = (result + 1) * 100
        result = pd.Series(
            result, index=pd.period_range(start_period, end_period, freq="M")
        )
        result.loc[start_period - 1] = 100.0
        result.sort_index(ascending=True, inplace=True)
        return result

    # Risk metrics

    @classmethod
    def get_portfolio_risk(cls, weights: Union[list, np.array], assets_ror: pd.DataFrame) -> float:
        """
        Compute the standard deviation of return for monthly rebalanced portfolio.
        """
        # cls.weights_sum_is_one(weights)
        if isinstance(assets_ror, pd.Series):  # required for a single asset portfolio
            return assets_ror.std()
        weights = np.array(weights)
        covmat = assets_ror.cov()
        return math.sqrt(weights.T @ covmat @ weights)

    @staticmethod
    def get_semideviation(
        ror: Union[pd.DataFrame, pd.Series]
    ) -> Union[pd.Series, float]:
        """
        Returns semideviation for each asset given returns time series.
        """
        is_negative = ror < 0
        return ror[is_negative].std(ddof=0)

    @staticmethod
    def get_var_historic(
        ror: Union[pd.DataFrame, pd.Series], level: int = 5
    ) -> Union[pd.Series, float]:
        """
        Compute monthly historic Value at Risk (VaR) at a specified level.
        """
        s = -ror.quantile(level / 100)
        if isinstance(s, pd.Series):
            s.name = "VaR"
        return s

    @staticmethod
    def get_cvar_historic(
        ror: Union[pd.DataFrame, pd.Series], level: int = 5
    ) -> Union[pd.Series, float]:
        """
        Compute the Conditional VaR (CVaR) at a specified level.
        """
        is_beyond = ror <= ror.quantile(
            level / 100
        )  # mask: return is less than quantile
        s = -ror[is_beyond].mean()
        if isinstance(s, pd.Series):
            s.name = "CVaR"
        return s

    @staticmethod
    def get_drawdowns(
        ror: Union[pd.DataFrame, pd.Series]
    ) -> Union[pd.DataFrame, pd.Series]:
        """
        Get drawdowns from return time series.
        """
        wealth_index = 1000 * (1 + ror).cumprod()
        previous_peaks = wealth_index.cummax()
        return (wealth_index - previous_peaks) / previous_peaks

    @staticmethod
    def change_columns_order(
        df: pd.DataFrame, selected_columns: list, position: str = "first"
    ) -> pd.DataFrame:
        """
        Places selected_columns on the first position (position='first') or last position (position='last').
        """
        cols = list(df.columns.values)  # Make a list of all of the columns in the df

        def condition(y):
            return y in selected_columns

        cols = [x for x in cols if not condition(x)]  # Remove from list
        # Create new DataFrame with columns in the right order
        if position == "first":
            df = df[selected_columns + cols]
        elif position == "last":
            df = df[cols + selected_columns]
        return df

    @staticmethod
    def skewness(ror: Union[pd.DataFrame, pd.Series]) -> Union[pd.Series, float]:
        """
        Calculate expanding skewness.
        The shape of time series should be at least 12. In the opposite case empty time series is returned.
        """
        # TODO: implement skewtest (from scipy)
        sk = ror.expanding(min_periods=1).skew()
        return sk.iloc[_MONTHS_PER_YEAR:]

    @staticmethod
    def skewness_rolling(
        ror: Union[pd.DataFrame, pd.Series], window: int = 60
    ) -> Union[pd.Series, float]:
        """
        Calculate rolling skewness.
        Window size should be at least 12 months.
        """
        check_rolling_window(window, ror)
        sk = ror.rolling(window=window).skew()
        sk.dropna(inplace=True)
        return sk

    @staticmethod
    def kurtosis(ror: Union[pd.Series, pd.DataFrame]):
        """
        Calculate expanding Fisher (normalized) kurtosis time series.
        Kurtosis should be close to zero for normal distribution.
        """
        kt = ror.expanding(min_periods=1).kurt()
        return kt.iloc[_MONTHS_PER_YEAR:]

    @staticmethod
    def kurtosis_rolling(ror: Union[pd.Series, pd.DataFrame], window: int = 60):
        """
        Calculate rolling Fisher (normalized) kurtosis time series.
        Kurtosis should be close to zero for normal distribution.
        Window should be at least 12 months.
        """
        check_rolling_window(window, ror)
        kt = ror.rolling(window=window).kurt()
        return kt.dropna()

    @staticmethod
    def jarque_bera_series(ror: pd.Series) -> dict:
        """
        Jarque-Bera goodness of fit test on a single time series (Pandas Series).
        The Jarque-Bera test shows whether the returns have the skewness and kurtosis matching a normal distribution.

        Returns:
            (The test statistic, The p-value for the hypothesis test)
            Low statistic numbers correspond to normal distribution.
        """
        result_tuple = scipy.stats.jarque_bera(ror)[0], scipy.stats.jarque_bera(ror)[1]
        return {"statistic": result_tuple[0], "p-value": result_tuple[1]}

    @staticmethod
    def jarque_bera_dataframe(ror: pd.DataFrame) -> pd.DataFrame:
        """
        Jarque-Bera goodness of fit test on time series in form of Pandas DataFrame.
        The Jarque-Bera test shows whether the returns have the skewness and kurtosis matching a normal distribution.

        Returns:
            (The test statistic, The p-value for the hypothesis test)
            Low statistic numbers correspond to normal distribution.
        """
        result = ror.apply(scipy.stats.jarque_bera, axis=0)
        return result.set_index(pd.Index(["statistic", "p-value"]))

    @staticmethod
    def kstest_series(ror: pd.Series, distr: str = "norm") -> dict:
        """
        Kolmogorov-Smirnov test goodness of fit test on a single time series (Pandas Series).

        Returns:
            {'statistics': The test statistic, 'p-value': The p-value for the hypothesis test}
        """
        if distr == "norm":
            kstest = scipy.stats.kstest(ror, distr, scipy.stats.norm.fit(ror))
        elif distr == "lognorm":
            kstest = scipy.stats.kstest(ror, distr, scipy.stats.lognorm.fit(ror))
        else:
            raise ValueError('distr should be "norm" (default) or "lognormal".')
        return {"statistic": kstest[0], "p-value": kstest[1]}

    @staticmethod
    def kstest_dataframe(ror: pd.DataFrame, distr: str = "norm") -> pd.DataFrame:
        """
        Kolmogorov-Smirnov test for goodness of fit test on time series in form of Pandas DataFrame.

        Returns:
            (The test statistic, The p-value for the hypothesis test)
        """
        test_dict = {}
        for label, content in ror.items():
            test_values = Frame.kstest_series(content, distr=distr)
            test_dict.update({label: test_values})
        return pd.DataFrame.from_dict(test_dict, orient="columns")


class Rebalance:
    """
    Methods for rebalancing portfolio.
    """

    @staticmethod
    def wealth_ts(
        weights: list, ror: pd.DataFrame, *, period: str = "year"
    ) -> pd.Series:
        """
        Calculate wealth index time series of rebalanced portfolio given returns time series of the assets.
        Default rebalancing period is a Year (end of year)
        For not rebalanced portfolio set Period to 'none'
        """
        # Frame.weights_sum_is_one(weights)
        initial_inv = 1000
        wealth_index = pd.Series(dtype="float64")
        if period == "none":  # Not rebalanced portfolio
            inv_period_spread = np.asarray(weights) * initial_inv
            assets_wealth_indexes = inv_period_spread * (1 + ror).cumprod()
            wealth_index = assets_wealth_indexes.sum(axis=1)
        else:
            for x in ror.resample("Y"):
                df = x[1]  # select ror part of the grouped data
                inv_period_spread = np.asarray(weights) * initial_inv  # rebalancing
                assets_wealth_indexes = inv_period_spread * (1 + df).cumprod()
                wealth_index_local = assets_wealth_indexes.sum(axis=1)
                wealth_index = pd.concat(
                    [wealth_index, wealth_index_local], verify_integrity=True, sort=True
                )
                initial_inv = wealth_index.iloc[-1]
        return wealth_index

    @staticmethod
    def assets_wealth_ts(
            weights: list, ror: pd.DataFrame, *, period: str = "year"
    ) -> pd.DataFrame:
        """
        Calculate ASSETS wealth indexes time series of rebalanced portfolio given returns time series of the assets.
        Default rebalancing period is a Year (end of year)
        For not rebalanced portfolio set Period to 'none'
        """
        # Frame.weights_sum_is_one(weights)
        initial_inv = 1000
        assets_wealth_indexes = pd.DataFrame(dtype="float64")
        wealth_index = pd.Series(dtype="float64")
        if period == "none":  # Not rebalanced portfolio
            inv_period_spread = np.asarray(weights) * initial_inv
            assets_wealth_indexes = inv_period_spread * (1 + ror).cumprod()
        else:
            for x in ror.resample("Y"):
                df = x[1]  # select ror part of the grouped data
                inv_period_spread = np.asarray(weights) * initial_inv  # rebalancing
                assets_wealth_indexes_local = inv_period_spread * (1 + df).cumprod()
                assets_wealth_indexes = pd.concat(
                    [assets_wealth_indexes, assets_wealth_indexes_local],
                    verify_integrity=True, sort=True
                )
                wealth_index_local = assets_wealth_indexes_local.sum(axis=1)
                wealth_index = pd.concat(
                    [wealth_index, wealth_index_local], verify_integrity=True, sort=True
                )
                initial_inv = wealth_index.iloc[-1]
        return assets_wealth_indexes

    @staticmethod
    def assets_weights_ts(
            weights: list, ror: pd.DataFrame, *, period: str = "year"
    ) -> pd.DataFrame:
        """
        Calculate assets weights monthly time series for rebalanced portfolio.

        Parameters
        ----------
        weights
        ror
        period

        Returns
        -------

        """
        assets_wealth_indexes = Rebalance.assets_wealth_ts(weights=weights, ror=ror, period=period)
        portfolio_wealth_index = Rebalance.wealth_ts(weights=weights, ror=ror, period=period)
        return assets_wealth_indexes.divide(portfolio_wealth_index, axis=0)

    @staticmethod
    def return_ts(
        weights: Union[list, np.ndarray], ror: pd.DataFrame, *, period: str = "year"
    ) -> pd.Series:
        """
        Return monthly rate of return time series of rebalanced portfolio given returns time series of the assets.
        Default rebalancing period is a Year (end of year)
        For not rebalanced portfolio set Period to 'none'
        """
        # define data of the first period
        first_date = ror.index[0]
        return_first_period = ror.iloc[0] @ weights

        wealth_index = Rebalance.wealth_ts(
            weights, ror, period=period
        )
        ror = wealth_index.pct_change()
        ror.loc[
            first_date
        ] = return_first_period  # replaces NaN with the first period return
        return ror


class Date:
    @staticmethod
    def subtract_years(dt: pd.Timestamp, years: int) -> pd.Timestamp:
        """
        Subtract N years (integer) from a date. Used for time series.
        First month is +1 (if today is August the series should start at September to give 12 months).
        """
        if isinstance(years, int):
            if dt.month == 12:
                new_dt = dt.replace(year=dt.year - years + 1, month=1)  # for December
            else:
                new_dt = dt.replace(year=dt.year - years, month=dt.month + 1)
        else:
            raise TypeError("The period should be integer")
        return new_dt


class Index:
    @staticmethod
    def tracking_difference(accumulated_return: pd.DataFrame) -> pd.DataFrame:
        """
        Returns tracking difference for a rate of return time series.
        Assets are compared with the index or another benchmark.
        Index should be in the first position (first column).
        """
        if accumulated_return.shape[1] < 2:
            raise ValueError(
                "At least 2 symbols should be provided to calculate Tracking Difference."
            )
        initial_value = accumulated_return.iloc[0]
        difference = (
            accumulated_return.subtract(accumulated_return.iloc[:, 0], axis=0)
            / initial_value
        )
        difference.drop(
            difference.columns[0], axis=1, inplace=True
        )  # drop the first column (stock index data)
        return difference

    @staticmethod
    def tracking_difference_annualized(tracking_diff: pd.DataFrame) -> pd.DataFrame:
        """
        Annualize the values of tracking difference time series.
        Annual values are available for periods of more than 12 months.
        Returns for less than 12 months can't be annualized.
        """
        pwr = 12 / (1.0 + np.arange(tracking_diff.shape[0]))
        y = abs(tracking_diff)
        diff = (y + 1.0).pow(pwr, axis=0) - 1.0
        diff = np.sign(tracking_diff) * diff
        return diff.iloc[
            _MONTHS_PER_YEAR - 1 :
        ]  # returns for the first 11 months can't be annualized

    @staticmethod
    def tracking_error(ror: pd.DataFrame) -> pd.DataFrame:
        """
        Returns tracking error for a rate of return time series.
        Assets are compared with the index or another benchmark.
        Index should be in the first position (first column).
        """
        if ror.shape[1] < 2:
            raise ValueError(
                "At least 2 symbols should be provided to calculate Tracking Error."
            )
        cumsum = ror.subtract(ror.iloc[:, 0], axis=0).pow(2, axis=0).cumsum()
        cumsum.drop(
            cumsum.columns[0], axis=1, inplace=True
        )  # drop the first column (stock index data)
        tracking_error = cumsum.divide((1.0 + np.arange(ror.shape[0])), axis=0).pow(
            0.5, axis=0
        )
        return tracking_error * np.sqrt(12)

    @staticmethod
    def cov_cor(ror: pd.DataFrame, fn: str) -> pd.DataFrame:
        """
        Returns the accumulated correlation or covariance time series.
        The period should be at least 12 months.
        """
        if ror.shape[1] < 2:
            raise ValueError("At least 2 symbols should be provided.")
        if fn not in ["cov", "corr"]:
            raise ValueError("fn should be corr or cov")
        cov_matrix_ts = getattr(ror.expanding(), fn)()
        cov_matrix_ts = cov_matrix_ts.drop(index=ror.columns[1:], level=1).droplevel(1)
        cov_matrix_ts.drop(columns=ror.columns[0], inplace=True)
        return cov_matrix_ts.iloc[_MONTHS_PER_YEAR:]

    @staticmethod
    def rolling_cov_cor(
        ror: pd.DataFrame, window: int = 60, fn: str = "corr"
    ) -> pd.DataFrame:
        """
        Returns the rolling correlation (or covariance) time series.
        The history period should be at least 12 months.
        """
        if ror.shape[1] < 2:
            raise ValueError("At least 2 symbols should be provided.")
        if fn not in ["cov", "corr"]:
            raise ValueError("fn should be corr or cov")
        check_rolling_window(window, ror)
        cov_matrix_ts = getattr(ror.rolling(window=window), fn)()
        cov_matrix_ts = cov_matrix_ts.drop(index=ror.columns[1:], level=1).droplevel(1)
        cov_matrix_ts.drop(columns=ror.columns[0], inplace=True)
        cov_matrix_ts.dropna(inplace=True)
        return cov_matrix_ts

    @staticmethod
    def beta(ror: pd.DataFrame) -> pd.DataFrame:
        """
        Calculate beta coefficient time series.
        Index (or benchmark) should be in the first position (first column).
        The period should be at least 12 months.
        """
        cov = Index.cov_cor(ror, fn="cov")
        var = ror.expanding().var().drop(columns=ror.columns[0])
        var = var[_MONTHS_PER_YEAR:]
        return cov / var



"""Portfolio ratios"""

from typing import Union

import pandas as pd


def get_sharpe_ratio(
    pf_return: Union[float, pd.Series],
    rf_return: float,
    std_deviation: Union[float, pd.Series],
) -> Union[float, pd.Series]:
    """
    Calculate Sharpe ratio.
    """
    return (pf_return - rf_return) / std_deviation



import pytest
from pytest import approx
from pytest import mark

import numpy as np
from numpy.testing import assert_allclose

import pandas as pd

import okama as ok


@mark.rebalance
@mark.frontier
def test_init_efficient_frontier_reb():
    with pytest.raises(ValueError, match=r'The number of symbols cannot be less than two'):
        ok.EfficientFrontierReb(assets=['MCFTR.INDX'])


def test_repr(init_efficient_frontier_reb):
    value = pd.Series(dict(
        symbols="[SPY.US, GLD.US]",
        currency="RUB",
        first_date="2019-01",
        last_date="2020-02",
        period_length="1 years, 2 months",
        rebalancing_period="year",
        inflation="RUB.INFL",
    ))
    assert repr(init_efficient_frontier_reb) == repr(value)


@mark.rebalance
@mark.frontier
def test_gmv_annual_weights(init_efficient_frontier_reb):
    assert_allclose(init_efficient_frontier_reb.gmv_annual_weights, np.array([0.765787, 0.234213]), rtol=1e-2, atol=1e-2)


@mark.rebalance
@mark.frontier
def test_gmv_annual_values(init_efficient_frontier_reb):
    assert init_efficient_frontier_reb.gmv_annual_values[0] == approx(0.09660054, rel=1e-2)


@mark.rebalance
@mark.frontier
def test_max_return(init_efficient_frontier_reb):
    assert init_efficient_frontier_reb.global_max_return_portfolio['CAGR'] == approx(0.14904342, rel=1e-2)


@mark.rebalance
@mark.frontier
def test_ef_points_reb(init_efficient_frontier_reb):
    assert init_efficient_frontier_reb.ef_points['CAGR'].iloc[1] == approx(0.14268245, rel=1e-2)



import requests
from urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter

from ..settings import default_ticker, default_namespace


class API:
    """
    Set of methods to get data from API.
    """
    # TODO: introduce 'from' & 'to' for dates.

    api_url = "http://185.63.191.70:5000"
    default_timeout = 5  # seconds

    endpoint_ror = "/api/ts/ror/"
    endpoint_symbol = "/api/symbol/"
    endpoint_search = "/api/search/"
    endpoint_live_price = "/api/live_price/"
    endpoint_adjusted_close = "/api/ts/adjusted_close/"
    endpoint_close = "/api/ts/close/"
    endpoint_dividends = "/api/ts/dividends/"
    endpoint_nav = "/api/ts/nav/"
    endpoint_macro = "/api/ts/macro/"
    # namespaces endpoints
    endpoint_namespaces = "/api/namespaces/"
    endpoint_assets_namespaces = "/api/assets_namespaces/"
    endpoint_macro_namespaces = "/api/macro_namespaces/"
    endpoint_no_dividends_namespaces = "/api/no_dividends_namespaces/"

    @classmethod
    def connect(
        cls,
        endpoint: str = endpoint_ror,
        symbol: str = default_ticker,
        first_date: str = "1900-01-01",
        last_date: str = "2100-01-01",
        period: str = "d",
    ) -> str:
        session = requests.session()
        retry_strategy = Retry(total=3,
                               backoff_factor=0.1,
                               status_forcelist=[429, 500, 502, 503, 504])
        adapter = HTTPAdapter(max_retries=retry_strategy)
        request_url = cls.api_url + endpoint + symbol
        params = {"first_date": first_date, "last_date": last_date, "period": period}
        session.mount("https://", adapter)
        session.mount("http://", adapter)
        try:
            r = session.get(request_url, params=params, verify=False, timeout=cls.default_timeout)
            r.raise_for_status()
        except requests.exceptions.HTTPError as errh:
            if r.status_code == 404:
                raise requests.exceptions.HTTPError(f"{symbol} is not found in the database.") from errh
            raise requests.exceptions.HTTPError(
                f"HTTP error fetching data for {symbol}:",
                r.status_code,
                r.reason,
                request_url,
            ) from errh
        return r.text

    @classmethod
    def get_ror(
        cls,
        symbol: str = default_ticker,
        first_date: str = "1900-01-01",
        last_date: str = "2100-01-01",
        period: str = "m",
    ):
        return cls.connect(
            endpoint=cls.endpoint_ror,
            symbol=symbol,
            first_date=first_date,
            last_date=last_date,
            period=period,
        )

    @classmethod
    def get_adjusted_close(
        cls,
        symbol: str = default_ticker,
        first_date: str = "1900-01-01",
        last_date: str = "2100-01-01",
        period: str = "m",
    ):
        return cls.connect(
            endpoint=cls.endpoint_adjusted_close,
            symbol=symbol,
            first_date=first_date,
            last_date=last_date,
            period=period,
        )

    @classmethod
    def get_close(
        cls,
        symbol: str = default_ticker,
        first_date: str = "1900-01-01",
        last_date: str = "2100-01-01",
        period: str = "m",
    ):
        return cls.connect(
            endpoint=cls.endpoint_close,
            symbol=symbol,
            first_date=first_date,
            last_date=last_date,
            period=period,
        )

    @classmethod
    def get_dividends(
        cls,
        symbol: str = default_ticker,
        first_date: str = "1900-01-01",
        last_date: str = "2100-01-01",
    ):
        return cls.connect(
            endpoint=cls.endpoint_dividends,
            symbol=symbol,
            first_date=first_date,
            last_date=last_date,
        )

    @classmethod
    def get_nav(
        cls,
        symbol: str = default_ticker,
        first_date: str = "1900-01-01",
        last_date: str = "2100-01-01",
        period: str = "m",
    ):
        return cls.connect(
            endpoint=cls.endpoint_nav,
            symbol=symbol,
            first_date=first_date,
            last_date=last_date,
            period=period,
        )

    @classmethod
    def get_macro(
        cls,
        symbol: str = default_ticker,
        first_date: str = "1900-01-01",
        last_date: str = "2100-01-01",
    ):
        """
        Get macro time series (monthly).
        """
        return cls.connect(
            endpoint=cls.endpoint_macro,
            symbol=symbol,
            first_date=first_date,
            last_date=last_date,
            period="m",
        )

    @classmethod
    def get_namespaces(cls):
        return cls.connect(endpoint=cls.endpoint_namespaces, symbol="")

    @classmethod
    def get_symbols_in_namespace(cls, namespace: str = default_namespace):
        return cls.connect(endpoint=cls.endpoint_namespaces, symbol=namespace)

    @classmethod
    def get_assets_namespaces(cls):
        return cls.connect(endpoint=cls.endpoint_assets_namespaces, symbol="")

    @classmethod
    def get_macro_namespaces(cls):
        return cls.connect(endpoint=cls.endpoint_macro_namespaces, symbol="")

    @classmethod
    def get_no_dividends_namespaces(cls):
        return cls.connect(endpoint=cls.endpoint_no_dividends_namespaces, symbol="")

    @classmethod
    def get_symbol_info(cls, symbol: str):
        return cls.connect(endpoint=cls.endpoint_symbol, symbol=symbol)

    @classmethod
    def search(cls, search_string: str):
        return cls.connect(endpoint=cls.endpoint_search, symbol=search_string)

    @classmethod
    def get_live_price(cls, symbol: str):
        return cls.connect(endpoint=cls.endpoint_live_price, symbol=symbol)



"""
Tests the validator functions
"""

import pytest

from okama.common.validators import validate_integer, validate_real


# validate_integer
def test_validate_integer_valid():
    validate_integer(
        "arg",
        10,
        min_value=0,
        max_value=20,
        inclusive=False,
        custom_min_message="custom min msg",
        custom_max_message="custom max msg",
    )


@pytest.mark.parametrize("object_value, exception", [(1.5, TypeError), (-1, ValueError), (100, ValueError)])
def test_validate_integer_error(object_value, exception):
    with pytest.raises(exception) as ex:
        validate_integer("arg", object_value, min_value=100, inclusive=False)
    assert "arg" in str(ex.value)


def test_validate_integer_min_custom_msg():
    with pytest.raises(ValueError) as ex:
        validate_integer("arg", 10, min_value=100, custom_min_message="custom")
    assert str(ex.value) == "custom"


def test_validate_integer_max_std_err_msg():
    with pytest.raises(ValueError) as ex:
        validate_integer("arg", 10, min_value=1, max_value=5)
    assert "arg" in str(ex.value)
    assert "5" in str(ex.value)


def test_validate_integer_max_custom_err_msg():
    with pytest.raises(ValueError) as ex:
        validate_integer(
            "arg", 10, min_value=1, max_value=5, custom_max_message="custom"
        )
    assert str(ex.value) == "custom"


def test_validate_integer_valid_inclusive():
    validate_integer(
        "arg",
        10,
        min_value=0,
        max_value=20,
        inclusive=True,
        custom_min_message="custom min msg",
        custom_max_message="custom max msg",
    )


def test_validate_integer_valid_inclusive_equal():
    validate_integer(
        "arg", 10, min_value=10, max_value=20, inclusive=True,
    )


def test_validate_integer_min_std_err_msg_inclusive():
    with pytest.raises(ValueError) as ex:
        validate_integer("arg", 10, min_value=11, inclusive=True)
    assert "arg" in str(ex.value)
    assert "11" in str(ex.value)


def test_validate_integer_max_std_err_msg_inclusive():
    with pytest.raises(ValueError) as ex:
        validate_integer("arg", 10, min_value=1, max_value=5, inclusive=True)
    assert "arg" in str(ex.value)
    assert "5" in str(ex.value)


# validate_real


def test_validate_real_valid():
    validate_real("number", 1)


def test_validate_real_type_error():
    with pytest.raises(TypeError):
        validate_real("arg", "not a number")



import pandas as pd
from pytest import mark
from pytest import approx


@mark.inflation
@mark.usefixtures('_init_inflation')
class TestInflation:

    def test_get_infl_rub_data(self):
        assert self.infl_rub.first_date == pd.to_datetime('1991-01')
        assert self.infl_rub.pl.years == 10
        assert self.infl_rub.pl.months == 1
        assert self.infl_rub.name == 'Russia Inflation Rate'
        assert self.infl_rub.type == 'inflation'
        assert self.infl_rub.cumulative_inflation[-1] == approx(19576.47386585591, rel=1e-4)
        assert self.infl_rub.purchasing_power_1000 == approx(0.05107911300773333, rel=1e-4)
        assert self.infl_rub.rolling_inflation[-1] == approx(0.2070533602100877, rel=1e-4)

    def test_get_infl_usd_data(self):
        assert self.infl_usd.first_date == pd.to_datetime('1913-02')
        assert self.infl_usd.pl.years == 10
        assert self.infl_usd.pl.months == 0
        assert self.infl_usd.name == 'US Inflation Rate'
        assert self.infl_usd.type == 'inflation'
        assert self.infl_usd.cumulative_inflation[-1] == approx(0.7145424753209466, rel=1e-4)
        assert self.infl_usd.purchasing_power_1000 == approx(583.2459763429362, rel=1e-4)
        assert self.infl_usd.rolling_inflation[-1] == approx(-0.005813765681402461, rel=1e-4)

    def test_get_infl_eur_data(self):
        assert self.infl_eur.first_date == pd.to_datetime('1996-02')
        assert self.infl_eur.pl.years == 10
        assert self.infl_eur.pl.years == 10
        assert self.infl_eur.name == 'EU Inflation Rate'
        assert self.infl_eur.type == 'inflation'
        assert self.infl_eur.cumulative_inflation[-1] == approx(0.20267532488218776, rel=1e-4)
        assert self.infl_eur.purchasing_power_1000 == approx(831.4796016106495, rel=1e-4)
        assert self.infl_eur.rolling_inflation[-1] == approx(0.02317927930197139, rel=1e-4)

    def test_describe(self):
        description = self.infl_rub.describe(years=[5])
        assert list(description.columns) == ['property', 'period', 'Russia Inflation Rate']
        assert description.loc[3, 'Russia Inflation Rate'] == approx(3.0414434004010245, rel=1e-4)
        assert description.loc[5, 'Russia Inflation Rate'] == approx(247.43634907784974, rel=1e-4)

    def test_annual_inflation_ts(self):
        assert self.infl_rub.annual_inflation_ts.iloc[-1] == approx(0.02760000000000007, rel=1e-4)


@mark.rates
@mark.usefixtures('_init_rates')
class TestRates:

    def test_rates_rub(self):
        assert self.rates_rub.name == 'Max deposit rates (RUB) in Russian banks'
        assert self.rates_rub.first_date == pd.to_datetime('2015-01')
        assert self.rates_rub.last_date == pd.to_datetime('2020-02')

    def test_okid(self):
        assert self.rates_rub.okid.sum() == approx(6376.2308059223915, rel=1e-4)



import json
from typing import Optional

import pandas as pd

from .api_methods import API
from .namespaces import symbols_in_namespace


def search(search_string: str, namespace: Optional[str] = None, response_format: str = 'frame') -> json:
    # search for string in a single namespace
    if namespace:
        df = symbols_in_namespace(namespace.upper())
        condition1 = df['name'].str.contains(search_string, case=False)
        condition2 = df['ticker'].str.contains(search_string, case=False)
        frame_response = df[condition1 | condition2]
        if response_format.lower() == 'frame':
            return frame_response
        elif response_format.lower() == 'json':
            return frame_response.to_json(orient='records')
        else:
            raise ValueError('response_format must be "json" or "frame"')
    # search for string in all namespaces
    string_response = API.search(search_string)
    json_response = json.loads(string_response)
    if response_format.lower() == 'frame':
        df = pd.DataFrame(json_response[1:], columns=json_response[0])
        return df
    elif response_format.lower() == 'json':
        return json_response
    else:
        raise ValueError('response_format must be "json" or "frame"')



from setuptools import setup, find_packages
from pathlib import Path

# read description from README.md and version from okama/__init__.py
(long_description, version) = (None, None)
with (Path(__file__).parent / "README.md").open(encoding="utf-8") as f:
    long_description = f.read()
with (Path(__file__).parent / "okama" / "__init__.py").open(encoding="utf-8") as f:
    for version_line in [l for l in f if l.startswith("__version__") and "=" in l]:
        version = version_line.split("=")[-1].strip().strip("'").strip('"')
if not (long_description and version):
    raise RuntimeError("Unable to get description and version string.")

setup(
    name="okama",
    version=version,
    license="MIT",
    description="Modern Portfolio Theory (MPT) Python package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Sergey Kikevich",
    author_email="sergey@rostsber.ru",
    url="https://okama.io/",
    download_url="https://github.com/mbk-dev/okama/archive/v0.81.tar.gz",
    keywords=["finance", "investments", "efficient frontier", "python", "optimization"],
    packages=find_packages(),
    package_data={"tests": ["*.csv"]},
    install_requires=[
        "pandas>=0.25.0",
        "numpy>=1.16.5",
        "scipy>=0.14.0",
        "matplotlib",
        "requests",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



import numpy as np
import pandas as pd
import pytest
from pytest import approx
from pytest import mark
from numpy.testing import assert_array_equal
from pandas.testing import assert_series_equal, assert_frame_equal

import okama as ok

from .conftest import data_folder


def test_initialization_failing():
    with pytest.raises(ValueError, match=r"Number of tickers \(2\) should be equal to the weights number \(3\)"):
        ok.Portfolio(assets=['MCFTR.INDX', 'MCFTR.INDX', 'RUB.FX'], weights=[.3, .3, .4])


def test_repr(portfolio_rebalanced_year):
    value = pd.Series(dict(
        symbol="pf1.PF",
        assets="[RUB.FX, MCFTR.INDX]",
        weights="[0.5, 0.5]",
        rebalancing_period="year",
        currency="RUB",
        inflation="RUB.INFL",
        first_date="2015-01",
        last_date="2020-01",
        period_length="5 years, 1 months",
    ))
    assert repr(portfolio_rebalanced_year) == repr(value)


def test_symbol_failing(portfolio_rebalanced_year):
    with pytest.raises(ValueError, match='portfolio symbol must be a string ending with ".PF" namespace.'):
        portfolio_rebalanced_year.symbol = 1
    with pytest.raises(ValueError, match='portfolio symbol must end with ".PF"'):
        portfolio_rebalanced_year.symbol = 'Not_a_good_symbol_for_portfolio.US'
    with pytest.raises(ValueError, match='portfolio text symbol should not have whitespace characters.'):
        portfolio_rebalanced_year.symbol = 'Not a good symbol for portfolio.PF'


def test_symbol_setter(portfolio_rebalanced_year):
    portfolio_rebalanced_year.symbol = 'portfolio_1.PF'
    assert portfolio_rebalanced_year.symbol == 'portfolio_1.PF'


def test_ror_rebalance(
    portfolio_rebalanced_year, portfolio_not_rebalanced
):
    print(f'portfolio_rebalanced_year={portfolio_rebalanced_year}')
    assert portfolio_rebalanced_year.ror[-2] == approx(0.01361, rel=1e-2)
    assert portfolio_not_rebalanced.ror[-1] == approx(0.01359, rel=1e-2)


def test_ror(portfolio_rebalanced_month):
    portfolio_sample = pd.read_pickle(data_folder / "portfolio.pkl")
    actual = portfolio_rebalanced_month.ror
    actual.rename("portfolio", inplace=True)
    assert_series_equal(actual, portfolio_sample)


def test_weights(portfolio_rebalanced_month):
    assert portfolio_rebalanced_month.weights == [0.5, 0.5]


def test_weights_ts_rebalanced_month(portfolio_rebalanced_month):
    assert portfolio_rebalanced_month.weights_ts["RUB.FX"].iloc[-1] == approx(0.5, rel=1e-2)


def test_weights_ts_rebalanced_year(portfolio_rebalanced_year, portfolio_not_rebalanced):
    assert portfolio_rebalanced_year.weights_ts["RUB.FX"].iloc[-2] == approx(0.3907, rel=1e-2)


def test_weights_ts_not_rebalanced(portfolio_not_rebalanced):
    assert portfolio_not_rebalanced.weights_ts["RUB.FX"].iloc[-1] == approx(0.2770, rel=1e-2)


def test_mean_return(portfolio_rebalanced_month):
    assert portfolio_rebalanced_month.mean_return_monthly == approx(0.010854, rel=1e-2)
    assert portfolio_rebalanced_month.mean_return_annual == approx(0.138324, rel=1e-2)


def test_real_mean_return(portfolio_rebalanced_month):
    assert portfolio_rebalanced_month.real_mean_return == approx(0.07827, rel=1e-2)


@mark.parametrize(
    "window, real, expected",
    [(1, True, 0.01100), (12, False, 0.12269), (12, True, 0.0961)],
)
def test_get_rolling_cumulative_return(portfolio_rebalanced_month, window, real, expected):
    assert portfolio_rebalanced_month.get_rolling_cumulative_return(window=window, real=real).iloc[-1, 0] == approx(expected, rel=1e-2)


def test_assets_close_monthly(portfolio_not_rebalanced):
    assert portfolio_not_rebalanced.assets_close_monthly.iloc[-1, 0] == 63.0359  # RUB.FX
    assert portfolio_not_rebalanced.assets_close_monthly.iloc[-1, 1] == 5245.6  # MCFTR.INDX


def test_close_monthly(portfolio_not_rebalanced):
    assert portfolio_not_rebalanced.close_monthly.iloc[-1] == approx(2022, rel=1e-2)


def test_number_of_securities(portfolio_not_rebalanced):
    assert portfolio_not_rebalanced.number_of_securities.iloc[-1, 0] == approx(8.88, rel=1e-2)  # RUB.FX
    assert portfolio_not_rebalanced.number_of_securities.iloc[-1, 1] == approx(0.2787, abs=1e-2)  # MCFTR.INDX


def test_dividends(portfolio_dividends):
    assert portfolio_dividends.dividends.iloc[-1] == approx(13.90, rel=1e-2)


def test_dividend_yield(portfolio_dividends):
    assert portfolio_dividends.dividend_yield.iloc[-1] == approx(
        0.03949, rel=1e-2
    )


def test_risk(portfolio_rebalanced_month):
    assert portfolio_rebalanced_month.risk_monthly == approx(0.035718, rel=1e-2)
    assert portfolio_rebalanced_month.risk_annual == approx(0.139814, rel=1e-2)


def test_get_var_historic(portfolio_rebalanced_month):
    assert portfolio_rebalanced_month.get_var_historic(time_frame=1, level=5) == approx(
        0.03815, rel=1e-2
    )
    assert portfolio_rebalanced_month.get_var_historic(time_frame=5, level=1) == approx(
        0.0969, rel=1e-2
    )


def test_get_cvar_historic(portfolio_rebalanced_month):
    assert portfolio_rebalanced_month.get_cvar_historic(
        time_frame=1, level=5
    ) == approx(0.05016, rel=1e-2)
    assert portfolio_rebalanced_month.get_cvar_historic(
        time_frame=5, level=1
    ) == approx(0.10762, rel=1e-2)


def test_drawdowns(portfolio_not_rebalanced):
    assert portfolio_not_rebalanced.drawdowns.min() == approx(-0.1265, rel=1e-2)


def test_recovery_period(portfolio_not_rebalanced):
    assert portfolio_not_rebalanced.recovery_period == 12


def test_get_cagr(portfolio_rebalanced_month, portfolio_no_inflation):
    values = pd.Series({"pf1.PF": 0.1303543, "RUB.INFL": 0.05548082428015655})
    actual = portfolio_rebalanced_month.get_cagr()
    assert_series_equal(actual, values, rtol=1e-4)
    # no inflation
    values = pd.Series({"pf1.PF": 0.1303543})
    actual = portfolio_no_inflation.get_cagr()
    assert_series_equal(actual, values, rtol=1e-4)
    # failing if wrong period
    with pytest.raises(TypeError):
        portfolio_rebalanced_month.get_cagr(period="one year")


cagr_testdata1 = [
    (1, 0.0961),
    (None, 0.0710),
]


@mark.parametrize("input_data, expected", cagr_testdata1, ids=["1 year", "full period"],)
def test_get_cagr_real(portfolio_rebalanced_month, input_data, expected):
    assert portfolio_rebalanced_month.get_cagr(period=input_data, real=True).values[0] == approx(expected, abs=1e-2)


def test_get_cagr_real_no_inflation_exception(portfolio_no_inflation):
    with pytest.raises(ValueError):
        portfolio_no_inflation.get_cagr(period=1, real=True)


@mark.parametrize(
    "period, real, expected",
    [("YTD", False, 0.01505), (1, False, 0.12269), (2, True, 0.2299)],
)
def test_cumulative_return(portfolio_rebalanced_month, period, real, expected):
    assert portfolio_rebalanced_month.get_cumulative_return(
        period=period, real=real
    ).iloc[0] == approx(expected, abs=1e-2)


cumulative_return_fail = [
    (1.5, False, TypeError),
    (-1, False, ValueError),
    (1, True, ValueError),
]


@mark.parametrize("period, real, exception", cumulative_return_fail)
def test_cumulative_return_error(portfolio_no_inflation, period, real, exception):
    with pytest.raises(exception):
        portfolio_no_inflation.get_cumulative_return(period=period, real=real)


def test_describe_inflation(portfolio_rebalanced_month):
    description = portfolio_rebalanced_month.describe()
    description_sample = pd.read_pickle(data_folder / "portfolio_description.pkl")
    assert_frame_equal(description, description_sample)


def test_describe_no_inflation(portfolio_no_inflation):
    description = portfolio_no_inflation.describe()
    description_sample = pd.read_pickle(data_folder / "portfolio_description_no_inflation.pkl")
    assert_frame_equal(description, description_sample)


def test_percentile_from_history(portfolio_rebalanced_month, portfolio_no_inflation, portfolio_short_history):
    assert portfolio_rebalanced_month.percentile_history_cagr(years=1).iloc[0, 1] == approx(0.12456, rel=1e-2)
    assert portfolio_no_inflation.percentile_history_cagr(years=1).iloc[0, 1] == approx(0.12456, rel=1e-2)
    with pytest.raises(
        ValueError,
        match="Time series does not have enough history to forecast. "
        "Period length is 0.90 years. At least 2 years are required.",
    ):
        portfolio_short_history.percentile_history_cagr(years=1)


def test_table(portfolio_rebalanced_month):
    assert_array_equal(
        portfolio_rebalanced_month.table["ticker"].values,
        np.array(["RUB.FX", "MCFTR.INDX"]),
    )


@mark.parametrize(
    "window, real, expected",
    [(12, False, 0.1290), (24, True, 0.08505)],
)
def test_get_rolling_cagr(portfolio_rebalanced_month, window, real, expected):
    assert portfolio_rebalanced_month.get_rolling_cagr(
        window=window, real=real).iloc[0, -1] == approx(expected, rel=1e-2)


def test_get_rolling_cagr_failing_short_window(portfolio_not_rebalanced):
    with pytest.raises(
        ValueError,
        match="window size should be at least 1 year"
    ):
        portfolio_not_rebalanced.get_rolling_cagr(window=1)


def test_get_rolling_cagr_failing_long_window(portfolio_not_rebalanced):
    with pytest.raises(
        ValueError,
        match="window size is more than data history depth"
    ):
        portfolio_not_rebalanced.get_rolling_cagr(window=100)


def test_get_rolling_cagr_failing_no_inflation(portfolio_no_inflation):
    with pytest.raises(
        ValueError,
        match="Real return is not defined. Set inflation=True when initiating the class."
    ):
        portfolio_no_inflation.get_rolling_cagr(real=True)


def test_monte_carlo_wealth(portfolio_rebalanced_month):
    assert portfolio_rebalanced_month._monte_carlo_wealth(
        distr='norm',
        years=1,
        n=1000
    ).iloc[-1, :].mean() == approx(2121, rel=1e-1)


@mark.parametrize(
    "distribution, expected",
    [('hist', 2096), ('norm', 2103), ('lognorm', 2093)],
)
def test_percentile_wealth(portfolio_rebalanced_month, distribution, expected):
    dic = portfolio_rebalanced_month.percentile_wealth(distr=distribution, years=1, n=100, percentiles=[50])
    assert dic[50] == approx(expected, rel=1e-1)


def test_forecast_monte_carlo_cagr(portfolio_rebalanced_month):
    dic = portfolio_rebalanced_month.percentile_distribution_cagr(years=2, distr='lognorm', n=100, percentiles=[50])
    assert dic[50] == approx(0.12, abs=5e-2)
    with pytest.raises(ValueError):
        portfolio_rebalanced_month.percentile_distribution_cagr(years=10, distr='lognorm', n=100, percentiles=[50])


def test_skewness(portfolio_rebalanced_month):
    assert portfolio_rebalanced_month.skewness.iloc[-1] == approx(2.39731, rel=1e-2)


def test_rolling_skewness(portfolio_rebalanced_month):
    assert portfolio_rebalanced_month.skewness_rolling(window=24).iloc[-1] == approx(
        0.82381, rel=1e-2
    )


def test_kurtosis(portfolio_rebalanced_month):
    assert portfolio_rebalanced_month.kurtosis.iloc[-1] == approx(13.19578, rel=1e-2)


def test_kurtosis_rolling(portfolio_rebalanced_month):
    assert portfolio_rebalanced_month.kurtosis_rolling(window=24).iloc[-1] == approx(
        1.58931, rel=1e-2
    )


def test_jarque_bera(portfolio_rebalanced_month):
    assert portfolio_rebalanced_month.jarque_bera["statistic"] == approx(
        424.1337, rel=1e-2
    )


def test_get_sharpe_ratio(portfolio_no_inflation):
    assert portfolio_no_inflation.get_sharpe_ratio(rf_return=0.05) == approx(0.631, rel=1e-2)


# This test should be a last one, as it changes the weights
def test_init_portfolio_failing():
    with pytest.raises(
        ValueError,
        match=r"Number of tickers \(2\) should be equal to the weights number \(3\)",
    ):
        ok.Portfolio(['RUB.FX', 'MCFTR.INDX'], weights=[0.1, 0.2, 0.7])
    with pytest.raises(ValueError, match="Weights sum is not equal to one."):
        ok.Portfolio(['RUB.FX', 'MCFTR.INDX'], weights=[0.1, 0.2])



"""
Tests the search
"""
import json

import pytest

from okama.api.search import search


def test_search_namespace_json():
    x = search(
        "aeroflot",
        namespace="MOEX",
        response_format="json",
    )
    assert json.loads(x)[0]['symbol'] == 'AFLT.MOEX'


def test_search_namespace_frame():
    x = search(
        "aeroflot",
        namespace="MOEX",
        response_format="frame",
    )
    assert x['symbol'].values[0] == 'AFLT.MOEX'


def test_search_all_json():
    x = search(
        "lkoh",
        response_format="json",
    )
    assert x[1][0] == 'LKOH.MOEX'


def test_search_all_frame():
    x = search(
        "lkoh",
        response_format="frame",
    )
    assert x['symbol'].iloc[0] == 'LKOH.MOEX'


def test_search_error():
    with pytest.raises(ValueError):
        search("arg", response_format='txt')



from random import randint
from typing import Optional, List, Dict, Union, Tuple

import numpy as np
import pandas as pd
import scipy.stats
from matplotlib import pyplot as plt

from .common.helpers.helpers import Frame, Rebalance, Float, Date
from .common.helpers import ratios
from .common.make_asset_list import ListMaker
from .common.validators import validate_real
from .settings import _MONTHS_PER_YEAR


class Portfolio(ListMaker):
    """
    Implementation of investment portfolio.

    Investments portfolio is a type of financial asset (same as stocks, ETF, mutual funds, currencies etc.).
    Arguments are similar to AssetList but additionally Portfolio has:

    - weights
    - rebalancing_period
    - symbol

    The rebalancing is the action of bringing the portfolio that has deviated away
    from original target asset allocation back into line. After rebalancing the portfolio assets
    have original weights.

    Parameters
    ----------
    assets : list, default None
        List of assets. Could include tickers or asset like objects (Asset, Portfolio).
        If None a single asset list with a default ticker is used.

    first_date : str, default None
        First date of monthly return time series.
        If None the first date is calculated automatically as the oldest available date for the listed assets.

    last_date : str, default None
        Last date of monthly return time series.
        If None the last date is calculated automatically as the newest available date for the listed assets.

    ccy : str, default 'USD'
        Base currency for the list of assets. All risk metrics and returns are adjusted to the base currency.

    inflation: bool, default True
        Defines whether to take inflation data into account in the calculations.
        Including inflation could limit available data (last_date, first_date)
        as the inflation data is usually published with a one-month delay.
        With inflation = False some properties like real return are not available.

    weights : list of float, default None
        List of assets weights.
        The weight of an asset is the percent of an investment portfolio that corresponds to the asset.
        If weights = None an equally weighted portfolio is created (all weights are equal).

    rebalancing_period : {'month', 'year', 'none'}, default 'month'
        Rebalancing period (rebalancing frequency) is predetermined time intervals when
        the investor rebalances the portfolio. If 'none' assets weights are not rebalanced.

    symbol : str, default None
        Text symbol of portfolio. It is similar to tickers but have a namespace information.
        Portfolio symbol must end with .PF (all_weather_portfolio.PF).
        If None a random symbol is generated (portfolio_7802.PF).
    """

    def __init__(
        self,
        assets: Optional[List[str]] = None,
        *,
        first_date: Optional[str] = None,
        last_date: Optional[str] = None,
        ccy: str = "USD",
        inflation: bool = True,
        weights: Optional[List[float]] = None,
        rebalancing_period: str = "month",
        symbol: str = None,
    ):
        super().__init__(
            assets,
            first_date=first_date,
            last_date=last_date,
            ccy=ccy,
            inflation=inflation,
        )
        self._weights = None
        self.weights = weights
        self.assets_weights = dict(zip(self.symbols, self.weights))
        self._rebalancing_period = None
        self.rebalancing_period = rebalancing_period
        self._symbol = symbol or f'portfolio_{randint(1000, 9999)}.PF'

    def __repr__(self):
        dic = {
            "symbol": self.symbol,
            "assets": self.symbols,
            "weights": self.weights,
            "rebalancing_period": self.rebalancing_period,
            "currency": self.currency,
            "inflation": self.inflation if hasattr(self, "inflation") else "None",
            "first_date": self.first_date.strftime("%Y-%m"),
            "last_date": self.last_date.strftime("%Y-%m"),
            "period_length": self._pl_txt,
        }
        return repr(pd.Series(dic))

    def _add_inflation(self):
        if hasattr(self, "inflation"):
            return pd.concat(
                [self.ror, self.inflation_ts], axis=1, join="inner", copy="false"
            )
        else:
            return self.ror

    @property
    def weights(self) -> Union[list, tuple]:
        """
        Get or set assets weights in portfolio.

        If not defined equal weights are used for each asset.

        Weights must be a list (or tuple) of float values.

        Returns
        -------
            Values for the weights of assets in portfolio.

        Examples
        --------
        >>> x = ok.Portfolio(['SPY.US', 'BND.US'])
        >>> x.weights
        [0.5, 0.5]
        """
        return self._weights

    @weights.setter
    def weights(self, weights: Optional[List[float]]):
        if weights is None:
            # Equally weighted portfolio
            n = len(self.symbols)  # number of assets
            weights = list(np.repeat(1 / n, n))
        else:
            [validate_real("weight", weight) for weight in weights]
            Frame.weights_sum_is_one(weights)
            if len(weights) != len(set(self.symbols)):
                raise ValueError(
                    f"Number of tickers ({len(set(self.symbols))}) should be equal "
                    f"to the weights number ({len(weights)})"
                )
        self._weights = weights

    @property
    def weights_ts(self) -> pd.DataFrame:
        """
        Calculate assets weights time series.

        The weights of assets in Portfolio are not constant if rebalancing_period is different from 'month'.

        Returns
        -------
        DataFrame
            Weights of assets time series.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> reb_period='none'  # The Portfolio is not rebalanced.
        >>> pf = ok.Portfolio(['SPY.US', 'AGG.US'], weights=[0.5, 0.5], rebalancing_period=reb_period)
        >>> pf.weights_ts.plot()
        >>> plt.show()

        The weights of assets time series will differ significantly if the portfolio rebalancing_period is 1 year.

        >>> pf.rebalancing_period = 'year'  # set a new rebalancing period
        >>> pf.weights_ts.plot()
        >>> plt.show()
        """
        if self.rebalancing_period != 'month':
            return Rebalance.assets_weights_ts(ror=self.assets_ror, period=self.rebalancing_period, weights=self.weights)
        values = np.tile(self.weights, (self.ror.shape[0], 1))
        return pd.DataFrame(values, index=self.ror.index, columns=self.symbols)

    @property
    def rebalancing_period(self) -> str:
        """
        Return rebalancing period of the portfolio.

        Rebalancing is the process by which an investor restores their portfolio to its target allocation
        by selling and buying assets. After rebalancing all the assets have original weights.

        Rebalancing period (rebalancing frequency) is predetermined time intervals when
        the investor rebalances the portfolio.

        Returns
        -------
        str
            Portfolio rebalancing period.
        """
        return self._rebalancing_period

    @rebalancing_period.setter
    def rebalancing_period(self, rebalancing_period: str):
        if rebalancing_period in {'none', 'month', 'year'}:
            self._rebalancing_period = rebalancing_period
        else:
            raise ValueError('rebalancing_period must be "year", "month" or "none"')

    @property
    def symbol(self) -> str:
        """
        Return a text symbol of portfolio.

        Symbols are similar to tickers but have a namespace information:

        * SPY.US is a symbol
        * SPY is a ticker

        Portfolios have '.PF' as a namespace.

        Returns
        -------
        str
            Text symbol of the portfolio.

        Examples
        --------
        >>> p = ok.Portflio()
        >>> p.symbol  # a randomly generated symbol will be shown
        'portfolio_5312.PF'

        >>> p.symbol = 'spy_portfolo.PF'  # The symbol can be customized after initialization

        New Portfolio can have a custom symbol.

        >>> p = ok.Portfolio(symbol='aggressive.PF')
        >>> p.symbol
        'aggressive.PF'
        """
        return self._symbol

    @symbol.setter
    def symbol(self, text_symbol: str):
        if isinstance(text_symbol, str) and '.' in text_symbol:
            if " " in text_symbol:
                raise ValueError('portfolio text symbol should not have whitespace characters.')
            namespace = text_symbol.split(".", 1)[-1]
            if namespace == 'PF':
                self._symbol = text_symbol
            else:
                raise ValueError('portfolio symbol must end with ".PF"')
        else:
            raise ValueError('portfolio symbol must be a string ending with ".PF" namespace.')

    @property
    def name(self) -> str:
        """
        Return text name of portfolio.

        For Portfolio name is equal to symbol.

        Returns
        -------
        str
            Text name of the portfolio.

        >>> p = ok.Portfolio()
        >>> p.name
        'portfolio_5312.PF'
        """
        return self.symbol

    @property
    def ror(self) -> pd.Series:
        """
        Calculate monthly rate of return time series for portfolio.

        Returns
        -------
        Series
            Rate of return monthly time series for portfolio.

        Examples
        --------
        >>> pf = ok.Portfolio(first_date='2020-01', last_date='2020-12')
        >>> pf.ror
        Date
        2020-01   -0.0004
        2020-02   -0.0792
        2020-03   -0.1249
        2020-04    0.1270
        2020-05    0.0476
        2020-06    0.0177
        2020-07    0.0589
        2020-08    0.0698
        2020-09   -0.0374
        2020-10   -0.0249
        2020-11    0.1088
        2020-12    0.0370
        Freq: M, Name: portfolio_4669.PF, dtype: float64

        >>> import matplotlib.pyplot as plt
        >>> pf.ror.plot(kind='bar')
        >>> plt.show()
        """
        if self.rebalancing_period == 'month':
            s = Frame.get_portfolio_return_ts(self.weights, self.assets_ror)
        else:
            s = Rebalance.return_ts(
                self.weights, self.assets_ror, period=self.rebalancing_period
            )
        return s.rename(self.symbol, inplace=True)

    @property
    def wealth_index(self) -> pd.DataFrame:
        """
        Calculate wealth index time series for the portfolio and accumulated inflation.

        Wealth index (Cumulative Wealth Index) is a time series that presents the value of portfolio over
        historical time period. Accumulated inflation time series is added if `inflation=True` in the Portfolio.

        Wealth index is obtained from the accumulated return multiplicated by the initial investments.
        That is: 1000 * (Acc_Return + 1)
        Initial investments are taken as 1000 units of the Portfolio base currency.

        Returns
        -------
            Time series of wealth index values for portfolio and accumulated inflation.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> x = ok.Portfolio(['SPY.US', 'BND.US'])
        >>> x.wealth_index.plot()
        >>> plt.show()
        """
        df = self._add_inflation()
        df = Frame.get_wealth_indexes(df)
        df = self._make_df_if_series(df)
        return df

    def _make_df_if_series(self, ts):
        if isinstance(ts, pd.Series):  # should always return a DataFrame
            ts = ts.to_frame()
            ts.rename({1: self.symbol}, axis="columns", inplace=True)
        return ts

    @property
    def wealth_index_with_assets(self) -> pd.DataFrame:
        """
        Calculate wealth index time series for the portfolio, all assets and accumulated inflation.

        Wealth index (Cumulative Wealth Index) is a time series that presents the value of portfolio over
        historical time period. Accumulated inflation time series is added if `inflation=True` in the Portfolio.

        Wealth index is obtained from the accumulated return multiplicated by the initial investments.
        That is: 1000 * (Acc_Return + 1)
        Initial investments are taken as 1000 units of the Portfolio base currency.

        Returns
        -------
        DataFrame
            Time series of wealth index values for portfolio, each asset and accumulated inflation.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(['VOO.US', 'GLD.US'], weights=[0.8, 0.2])
        >>> pf.wealth_index_with_assets.plot()
        >>> plt.show()
        """
        if hasattr(self, "inflation"):
            df = pd.concat(
                [self.ror, self.assets_ror, self.inflation_ts],
                axis=1,
                join="inner",
                copy="false",
            )
        else:
            df = pd.concat(
                [self.ror, self.assets_ror], axis=1, join="inner", copy="false"
            )
        return Frame.get_wealth_indexes(df)

    @property
    def mean_return_monthly(self) -> float:
        """
        Calculate monthly mean return (arithmetic mean) for the portfolio rate of return time series.

        Mean return calculated for the full history period.

        Returns
        -------
        Float
            Mean return value.

        Examples
        --------
        >>> pf = ok.Portfolio(['ISF.LSE', 'XGLE.LSE'], weights=[0.6, 0.4], ccy='GBP')
        >>> pf
        0.0001803312727272665
        """
        return Frame.get_portfolio_mean_return(self.weights, self.assets_ror)

    @property
    def mean_return_annual(self) -> float:
        """
        Calculate annualized mean return (arithmetic mean) for the portfolio rate of return time series.

        Mean return calculated for the full history period.

        Returns
        -------
        Float
           Mean return value.

        Examples
        --------
        >>> pf = ok.Portfolio(['XCS6.XETR', 'PHAU.LSE'], weights=[0.85, 0.15], ccy='USD')
        >>> pf.names
        {'XCS6.XETR': 'Xtrackers MSCI China UCITS ETF 1C', 'PHAU.LSE': 'WisdomTree Physical Gold'}
        >>> pf.mean_return_annual
        0.09005826844072184
        """
        return Float.annualize_return(self.mean_return_monthly)

    @property
    def annual_return_ts(self) -> pd.Series:
        """
        Calculate annual rate of return time series for portfolio.

        Rate of return is calculated for each calendar year.

        Returns
        -------
        DataFrame
            Calendar annual rate of return time series.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(['VOO.US', 'AGG.US'], weights=[0.4, 0.6])
        >>> pf.annual_return_ts.plot(kind='bar')
        >>> plt.show()

        Plot annual returns for portfolio with EUR as the base currency.

        >>> pf = ok.Portfolio(['VOO.US', 'AGG.US'], weights=[0.4, 0.6], ccy='EUR')
        >>> pf.annual_return_ts.plot(kind='bar')
        >>> plt.show()
        """
        return Frame.get_annual_return_ts_from_monthly(self.ror)

    def get_cagr(self, period: Optional[int] = None, real: bool = False) -> pd.Series:
        """
        Calculate portfolio Compound Annual Growth Rate (CAGR) for a given trailing period.

        Compound annual growth rate (CAGR) is the rate of return that would be required for an investment to grow from
        its initial to its final value, assuming all incomes were reinvested.

        Inflation adjusted annualized returns (real CAGR) are shown with `real=True` option.

        Annual inflation value is calculated for the same period if inflation=True in the AssetList.

        Parameters
        ----------
        period: int, optional
            CAGR trailing period in years. None for the full time CAGR.
        real: bool, default False
            CAGR is adjusted for inflation (real CAGR) if True.
            Portfolio should be initiated with Inflation=True for real CAGR.

        Returns
        -------
        Series
            Portfolio CAGR value and annualized inflation (optional).

        Notes
        -----
        CAGR is not defined for periods less than 1 year (NaN values are returned).

        Examples
        --------
        >>> pf = ok.Portfolio(['XCS6.XETR', 'PHAU.LSE'], weights=[0.85, 0.15], ccy='USD')
        >>> pf.names
        {'XCS6.XETR': 'Xtrackers MSCI China UCITS ETF 1C', 'PHAU.LSE': 'WisdomTree Physical Gold'}

        To get inflation adjusted return (real annualized return) add `real=True` option:

        >>> pf.get_cagr(period=5, real=True)
        portfolio_5625.PF    0.121265
        dtype: float64
        """
        ts = self._add_inflation()
        df = self._make_df_if_series(ts)
        dt0 = self.last_date
        if period is None:
            dt = self.first_date
        else:
            self._validate_period(period)
            dt = Date.subtract_years(dt0, period)
        cagr = Frame.get_cagr(df[dt:])
        if real:
            if not hasattr(self, "inflation"):
                raise ValueError(
                    "Real CAGR is not defined. Set inflation=True in Portfolio to calculate it."
                )
            mean_inflation = Frame.get_cagr(self.inflation_ts[dt:])
            cagr = (1. + cagr) / (1. + mean_inflation) - 1.
            cagr.drop(self.inflation, inplace=True)
        return cagr

    def get_rolling_cagr(self, window: int = 12, real: bool = False) -> pd.DataFrame:
        """
        Calculate rolling CAGR (Compound Annual Growth Rate) for the portfolio.

        Parameters
        ----------
        window : int, default 12
            Size of the moving window in months. Window size should be at least 12 months for CAGR.
        real: bool, default False
            CAGR is adjusted for inflation (real CAGR) if True.
            Portfolio should be initiated with Inflation=True for real CAGR.

        Returns
        -------
        DataFrame
            Time series of rolling CAGR and mean inflation (optionaly).

        Notes
        -----
        CAGR is not defined for periods less than 1 year (NaN values are returned).

        Examples
        --------
        >>> x = ok.Portfolio(['DXET.XETR', 'DBXN.XETR'], ccy='EUR', inflation=True)
        >>> x.get_rolling_cagr(window=5*12, real=True)
                 portfolio_7645.PF
        2013-09           0.029914
        2013-10           0.052435
        2013-11           0.055651
        2013-12           0.045180
        2014-01           0.063153
                            ...
        2021-01           0.032734
        2021-02           0.037779
        2021-03           0.043811
        2021-04           0.043729
        2021-05           0.042704
        """
        df_or_ts = self._add_inflation()
        if real:
            df_or_ts = self._make_real_return_time_series(df_or_ts)
        df = self._make_df_if_series(df_or_ts)
        return Frame.get_rolling_fn(df, window=window, fn=Frame.get_cagr)

    def get_cumulative_return(self, period: Union[str, int, None] = None, real: bool = False) -> pd.Series:
        """
        Calculate cumulative return over a given trailing period for the portfolio.

        The cumulative return is the total change in the portfolio price during the investment period.

        Inflation adjusted cumulative returns (real cumulative returns) are shown with `real=True` option.
        Annual inflation data is calculated for the same period if `inflation=True` in the AssetList.

        Parameters
        ----------
        period: str, int or None, default None
            Trailing period in years.
            None - full time cumulative return.
            'YTD' - (Year To Date) period of time beginning the first day of the calendar year up to the last month.
        real: bool, default False
            Cumulative return is adjusted for inflation (real cumulative return) if True.
            Portfolio should be initiated with `Inflation=True` for real cumulative return.

        Returns
        -------
        Series
            Cumulative rate of return values for portfolio and cumulative inflation (if inflation=True in Portfolio).

        Examples
        --------
        >>> pf = ok.Portfolio(['BTC-USD.CC', 'LTC-USD.CC'], weights=[.8, .2], last_date='2021-03')
        >>> pf.get_cumulative_return(period=2, real=True)
        portfolio_6232.PF    9.39381
        dtype: float64
        """
        ts = self._add_inflation()
        df = self._make_df_if_series(ts)
        dt0 = self.last_date

        if period is None:
            dt = self.first_date
        elif str(period).lower() == "ytd":
            year = dt0.year
            dt = str(year)
        else:
            self._validate_period(period)
            dt = Date.subtract_years(dt0, period)

        cr = Frame.get_cumulative_return(df[dt:])
        if real:
            if not hasattr(self, "inflation"):
                raise ValueError(
                    "Real cumulative return is not defined (no inflation information is available)."
                    "Set inflation=True in Portfolio to calculate it."
                )
            cumulative_inflation = Frame.get_cumulative_return(self.inflation_ts[dt:])
            cr = (1. + cr) / (1. + cumulative_inflation) - 1.
            cr.drop(self.inflation, inplace=True)
        return cr

    def get_rolling_cumulative_return(self, window: int = 12, real: bool = False) -> pd.DataFrame:
        """
        Calculate rolling cumulative return.

        The cumulative return is the total change in the portfolio price.

        Parameters
        ----------
        window : int, default 12
            Size of the moving window in months.
        real: bool, default False
            Cumulative return is adjusted for inflation (real cumulative return) if True.
            Portfolio should be initiated with `Inflation=True` for real cumulative return.

        Returns
        -------
        DataFrame
            Time series of rolling cumulative return and inflation (optional).

        Examples
        --------
        >>> pf = ok.Portfolio(['SPY.US', 'AGG.US', 'GLD.US'], weights=[.6, .35, .05], rebalancing_period='year')
        >>> pf.get_rolling_cumulative_return(window=24, real=True)
                 portfolio_9012.PF
        2006-11           0.125728
        2006-12           0.104348
        2007-01           0.129601
        2007-02           0.110680
        2007-03           0.132610
                            ...
        2021-03           0.263755
        2021-04           0.275474
        2021-05           0.322736
        2021-06           0.264963
        2021-07           0.273801
        [177 rows x 1 columns]
        """
        ts = self._add_inflation()
        if real:
            ts = self._make_real_return_time_series(ts)
        df = self._make_df_if_series(ts)
        return Frame.get_rolling_fn(
            df,
            window=window,
            fn=Frame.get_cumulative_return,
            window_below_year=True,
        )

    @property
    def assets_close_monthly(self) -> pd.DataFrame:
        """
        Show assets monthly close time series adjusted to the base currency.

        Returns
        -------
        DataFrame
            Assets monthly close time series adjusted to the base currency.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(['SPY.US', 'BND.US'], ccy='USD')
        >>> pf.assets_close_monthly.plot()
        >>> plt.show()
        """
        assets_close_monthly = pd.DataFrame(dtype=float)
        for i, x in enumerate(self.asset_obj_dict.values()):
            if i == 0:  # required to use pd.concat below (df should not be empty).
                assets_close_monthly = x.close_monthly if x.currency == self.currency else self._adjust_price_to_currency_monthly(x.close_monthly, x.currency)
                assets_close_monthly.rename(x.symbol, inplace=True)
            else:
                new = x.close_monthly if x.currency == self.currency else self._adjust_price_to_currency_monthly(x.close_monthly, x.currency)
                new.rename(x.symbol, inplace=True)
                assets_close_monthly = pd.concat([assets_close_monthly, new], axis=1, join="inner", copy="false")
        if isinstance(assets_close_monthly, pd.Series):
            assets_close_monthly = assets_close_monthly.to_frame()
        assets_close_monthly = assets_close_monthly[self.first_date: self.last_date]
        return assets_close_monthly

    @property
    def close_monthly(self) -> pd.Series:
        """
        Portfolio size monthly time series.

        Portfolio size is shown in base currency units. It is similar to the close value of an asset.
        Initial portfolio value is equal to 1000 units of base currency.

        Returns
        -------
        pd.Series
            Monthly portfolio size time series.

        Notes
        -----
        'close_mothly' shows the same output as the 'wealth_index'.
        This property is required as Portfolio must have the same attributes as an Asset.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(['SPY.US', 'BND.US'], ccy='USD')
        >>> pf.close_monthly.plot()
        >>> plt.show()
        """
        return self.wealth_index.iloc[:, 0]

    @property
    def number_of_securities(self) -> pd.DataFrame:
        """
        Calculate the number of securities monthly time series for the portfolio assets.

        Number of securities in the Portfolio is changing over time as the dividends are reinvested.
        Portfolio rebalancing also affects the number of securities.

        Initial number of securities depends on the portfolio size in base currency (1000 units).

        Returns
        -------
        DataFrame
            Number of securities monthly time series for the portfolio assets.

        Examples
        --------
        >>> pf = ok.Portfolio(['SPY.US', 'BND.US'], ccy='USD', last_date='07-2021')
        >>> pf.number_of_securities
                   SPY.US     BND.US
        Date
        2007-05  3.261153   6.687174
        2007-06  3.337216   6.758447
        2007-07  3.407015   6.643519
        2007-08  3.410268   6.663862
        2007-09  3.372630   6.798730
                   ...        ...
        2021-03  3.273521  15.313911
        2021-04  3.204779  15.685601
        2021-05  3.196768  15.749127
        2021-06  3.186124  15.879056
        2021-07  3.166335  16.003569
        [171 rows x 2 columns]
        """
        return self.weights_ts.mul(self.wealth_index.iloc[:, 0], axis=0).div(self.assets_close_monthly, axis=0)

    @property
    def dividends(self) -> pd.Series:
        """
        Calculate portfolio dividends monthly time series.

        Portfolio dividends are obtained by summing asset dividends adjusted to the base currency.
        Dividends size depends on the portfolio value and number of securities.

        Returns
        -------
        Series
            Portfolio dividends monthly time series.

        Examples
        --------
        >>> pf = ok.Portfolio(['SPY.US', 'BND.US'], ccy='USD', last_date='07-2021')
        >>> pf.dividends
        2007-05    0.849271
        2007-06    3.928855
        2007-07    1.551262
        2007-08    2.023148
        2007-09    4.423416
                     ...
        2021-03    6.155337
        2021-04    3.019478
        2021-05    2.056836
        2021-06    6.519521
        2021-07    2.114071
        Freq: M, Name: portfolio_2951.PF, Length: 171, dtype: float64
        """
        s = (self._get_assets_dividends() * self.number_of_securities).sum(axis=1)
        s.rename(self.symbol, inplace=True)
        return s

    @property
    def dividend_yield(self) -> pd.Series:
        """
        Calculate last twelve months (LTM) dividend yield time series for the portfolio. Time series has monthly values.

        Portfolio dividend yield is a weighted sum of the assets dividend yields (adjusted to
        the portfolio base currency).

        For an asset LTM dividend yield is the sum trailing twelve months of common dividends per share divided by
        the current price per share.

        Returns
        -------
        Series
            Portfolio LTM dividend yield monthly time series.

        Examples
        --------
        >>> pf = ok.Portfolio(['T.US', 'XOM.US'], weights=[0.8, 0.2], first_date='2010-01', last_date='2021-01', ccy='USD')
        >>> pf.dividend_yield
        2010-01    0.013249
        2010-02    0.014835
        2010-03    0.014257
                     ...
        2020-11    0.076132
        2020-12    0.074743
        2021-01    0.073643
        Freq: M, Name: portfolio_8836.PF, Length: 133, dtype: float64

        >>> import matplotlib.pyplot as plt
        >>> pf.dividend_yield.plot()
        >>> plt.show()
        """
        df = self.assets_dividend_yield @ self.weights_ts.T
        div_yield_series = pd.Series(np.diag(df), index=df.index)
        div_yield_series.rename(self.symbol, inplace=True)
        return div_yield_series

    @property
    def real_mean_return(self) -> float:
        """
        Calculate annualized real mean return (arithmetic mean) for the rate of return time series.

        Real rate of return is adjusted for inflation. Real return is defined if
        there is an `inflation=True` option in Portfolio.

        Returns
        -------
        float
            Annualized value of the mean for the real rate of return time series.

        Examples
        --------
        >>> pf = ok.Portfolio(['MSFT.US', 'AAPL.US'])
        >>> pf.real_mean_return
        0.3088967455111862
        """
        if not hasattr(self, "inflation"):
            raise ValueError(
                "Real Return is not defined. Set inflation=True to calculate."
            )
        infl_mean = Float.annualize_return(self.inflation_ts.mean())
        ror_mean = Float.annualize_return(self.ror.mean())
        return (1.0 + ror_mean) / (1.0 + infl_mean) - 1.0

    @property
    def risk_monthly(self) -> float:
        """
        Calculate monthly risk (standard deviation of return) for Portfolio.

        Monthly risk of portfolio is a standard deviation of the rate of return time series.
        Standard deviation (sigma σ) is normalized by N-1.

        Returns
        -------
        float
            Standard deviation value of the monthly return time series.

        See Also
        --------
        risk_annual : Calculate annualized risks.
        semideviation_monthly : Calculate semideviation monthly values.
        semideviation_annual : Calculate semideviation annualized values.
        get_var_historic : Calculate historic Value at Risk (VaR).
        get_cvar_historic : Calculate historic Conditional Value at Risk (CVaR).
        drawdowns : Calculate drawdowns.

        Examples
        --------
        >>> pf = ok.Portfolio(['MSFT.US', 'AAPL.US'])
        >>> pf.risk_monthly
        0.09415483565833212
        """
        return self.ror.std()

    @property
    def risk_annual(self) -> float:
        """
        Calculate annualized risk (return standard deviation) for portfolio.

        Returns
        -------
        float
            Annualized standard deviation value of the monthly return time series.

        Examples
        --------
        >>> pf = ok.Portfolio(['MSFT.US', 'AAPL.US'])
        >>> pf.risk_annual
        0.4374591902169046
        """
        return Float.annualize_risk(self.risk_monthly, self.mean_return_monthly)

    @property
    def semideviation_monthly(self) -> float:
        """
        Calculate semi-deviation monthly value for portfolio rate of return time series.

        Semi-deviation (Downside risk) is the risk of the return being below the expected return.

        Returns
        -------
        float
            Semi-deviation monthly value for portfolio rate of return time series.

        Examples
        --------
        >>> pf = ok.Portfolio(['MSFT.US', 'AAPL.US'])
        >>> pf.semideviation_monthly
        0.05601433676604449
        """
        return Frame.get_semideviation(self.ror)

    @property
    def semideviation_annual(self) -> float:
        """
        Return semideviation annualized value for portfolio rate of return time series.

        Semi-deviation (Downside risk) is the risk of the return being below the expected return.

        Returns
        -------
        float
            Annualized semi-deviation monthly value for portfolio rate of return time series.

        Examples
        --------
        >>> pf = ok.Portfolio(['MSFT.US', 'AAPL.US'])
        >>> pf.semideviation_annual
        0.1940393544621248
        """
        return Frame.get_semideviation(self.ror) * 12 ** 0.5

    def get_var_historic(self, time_frame: int = 12, level=1) -> float:
        """
        Calculate historic Value at Risk (VaR) for the portfolio.

        The VaR calculates the potential loss of an investment with a given time frame and confidence level.
        Loss is a positive number (expressed in cumulative return).
        If VaR is negative there are expected gains at this confidence level.

        Parameters
        ----------
        time_frame : int, default 12
            Time frame for VAR. Default is 12 months.
        level : int, default 1
            Confidence level in percents. Default value is 1%.

        Returns
        -------
        Float
            Historic Value at Risk (VaR) value for the portfolio.

        Examples
        --------
        >>> x = ok.Portfolio(['SP500TR.INDX', 'SP500BDT.INDX'], last_date='2021-01')
        >>> x.get_var_historic(time_frame=12, level=1)
        0.24030006476701732
        """
        # remove inflation column from rolling return
        df = self.get_rolling_cumulative_return(window=time_frame).loc[:, [self.symbol]]
        return Frame.get_var_historic(df, level).iloc[0]

    def get_cvar_historic(self, time_frame: int = 12, level=1) -> float:
        """
        Calculate historic Conditional Value at Risk (CVAR, expected shortfall) for the portfolio.

        CVaR is the average loss over a specified time period of unlikely scenarios beyond the confidence level.
        Loss is a positive number (expressed in cumulative return).
        If CVaR is negative there are expected gains at this confidence level.

        Parameters
        ----------
        time_frame : int, default 12 (12 months)
        level : int, default 1 (1% quantile)

        Returns
        -------
        Float
            Historic Conditional Value at Risk (CVAR, expected shortfall) value for the portfolio.

        Examples
        --------
        >>> x = ok.Portfolio(['USDEUR.FX', 'BTC-USD.CC'], last_date='2021-01')
        >>> x.get_cvar_historic(time_frame=2, level=1)
        0.3566909250442616
        """
        # remove inflation column form rolling return
        df = self.get_rolling_cumulative_return(window=time_frame).loc[:, [self.symbol]]
        return Frame.get_cvar_historic(df, level).iloc[0]

    @property
    def drawdowns(self) -> pd.Series:
        """
        Calculate drawdowns time series for the portfolio.

        The drawdown is the percent decline from a previous peak in wealth index.

        Returns
        -------
        Series
            Drawdowns time series for the portfolio
        """
        return Frame.get_drawdowns(self.ror)

    @property
    def recovery_period(self) -> int:
        """
        Calculate the longest recovery period for the portfolio assets value.

        The recovery period (drawdown duration) is the number of months to reach the value of the last maximum.

        Returns
        -------
        Integer
            Max recovery period for the protfolio assets value in months.

        Notes
        -----
        If the last maximum value is not recovered NaN is returned.
        The largest recovery period does not necessary correspond to the max drawdown.

        Examples
        --------
        >>> pf = ok.Portfolio(['SPY.US', 'AGG.US'], weights=[0.5, 0.5])
        >>> pf.recovery_period
        35

        See Also
        --------
        drawdowns : Calculate drawdowns time series.
        """
        if hasattr(self, "inflation"):
            w_index = self.wealth_index.drop(columns=[self.inflation])
        else:
            w_index = self.wealth_index
        if isinstance(w_index, pd.DataFrame):
            # time series should be a Series to use groupby
            w_index = w_index.squeeze()
        cummax = w_index.cummax()
        s = cummax.pct_change()[1:]
        s1 = s.where(s == 0).notnull().astype(int)
        s1_1 = s.where(s == 0).isnull().astype(int).cumsum()
        s2 = s1.groupby(s1_1).cumsum()
        # Max recovery period date should not be in the border (means it's not recovered)
        max_period = s2.max() if s2.idxmax().to_timestamp() != self.last_date else np.NAN
        return max_period

    def describe(self, years: Tuple[int] = (1, 5, 10)) -> pd.DataFrame:
        """
        Generate descriptive statistics for the portfolio.

        Statistics includes:

        - YTD (Year To date) compound return
        - CAGR for a given list of periods
        - LTM Dividend yield - last twelve months dividend yield

        Risk metrics (full available period):

        - risk (standard deviation)
        - CVAR (timeframe is 1 year)
        - max drawdowns (and dates)

        Parameters
        ----------
        years : tuple of (int,), default (1, 5, 10)
            List of periods for CAGR statistics.

        Returns
        -------
        DataFrame
            Table of descriptive statistics for the portfolio.

        See Also
        --------
            get_cumulative_return : Calculate cumulative return.
            get_cagr : Calculate assets Compound Annual Growth Rate (CAGR).
            dividend_yield : Calculate dividend yield (LTM).
            risk_annual : Return annualized risks (standard deviation).
            get_cvar : Calculate historic Conditional Value at Risk (CVAR, expected shortfall).
            drawdowns : Calculate drawdowns.

        Examples
        --------
        >>> pf = ok.Portfolio(['SPY.US', 'BND.US'], ccy='USD', last_date='07-2021')
        >>> pf.describe(years=[2, 5, 7])  # 'years' customizes the timeframe for the CAGR
                    property              period portfolio_2951.PF  inflation
        0    compound return                 YTD          0.084098   0.048154
        1               CAGR             2 years          0.141465   0.031566
        2               CAGR             5 years          0.102494   0.025582
        3               CAGR             7 years          0.091694   0.019656
        4               CAGR  14 years, 3 months          0.074305   0.019724
        5     Dividend yield                 LTM          0.016504        NaN
        6               Risk  14 years, 3 months          0.086103        NaN
        7               CVAR  14 years, 3 months          0.214207        NaN
        8       Max drawdown  14 years, 3 months         -0.266915        NaN
        9  Max drawdown date  14 years, 3 months           2009-02        NaN
        """
        description = pd.DataFrame()
        dt0 = self.last_date
        df = self._add_inflation()
        # YTD return
        ytd_return = self.get_cumulative_return(period="YTD")
        row = ytd_return.to_dict()
        row.update(period="YTD", property="compound return")
        description = description.append(row, ignore_index=True)
        # CAGR for a list of periods
        if self.pl.years >= 1:
            for i in years:
                dt = Date.subtract_years(dt0, i)
                if dt >= self.first_date:
                    row = self.get_cagr(period=i).to_dict()
                else:
                    row = (
                        {x: None for x in df.columns}
                        if hasattr(self, "inflation")
                        else {self.symbol: None}
                    )
                row.update(period=f"{i} years", property="CAGR")
                description = description.append(row, ignore_index=True)
            # CAGR for full period
            row = self.get_cagr(period=None).to_dict()
            row.update(period=self._pl_txt, property="CAGR",)
            description = description.append(row, ignore_index=True)
            # Dividend Yield
            value = self.dividend_yield.iloc[-1]
            row = {self.symbol: value}
            row.update(period="LTM", property=f"Dividend yield",)
            description = description.append(row, ignore_index=True)
        # risk (standard deviation)
        row = {self.symbol: self.risk_annual}
        row.update(
            period=self._pl_txt, property="Risk"
        )
        description = description.append(row, ignore_index=True)
        # CVAR
        if self.pl.years >= 1:
            row = {self.symbol: self.get_cvar_historic()}
            row.update(
                period=self._pl_txt,
                property="CVAR",
            )
            description = description.append(row, ignore_index=True)
        # max drawdowns
        row = {self.symbol: self.drawdowns.min()}
        row.update(
            period=self._pl_txt,
            property="Max drawdown",
        )
        description = description.append(row, ignore_index=True)
        # max drawdowns dates
        row = {self.symbol: self.drawdowns.idxmin()}
        row.update(
            period=self._pl_txt,
            property="Max drawdown date",
        )
        description = description.append(row, ignore_index=True)
        if hasattr(self, "inflation"):
            description.rename(columns={self.inflation: "inflation"}, inplace=True)
        description = Frame.change_columns_order(
            description, ["property", "period", self.symbol]
        )
        return description

    @property
    def table(self) -> pd.DataFrame:
        """
        Return table with security name, ticker, weight for assets in the portfolio.

        Returns
        -------
        DataFrame
            Security name - ticker - weight table.

        Examples
        --------
        >>> pf = ok.Portfolio(['MSFT.US', 'AAPL.US'])
        >>> pf.table
                        asset name   ticker  weights
        0  Microsoft Corporation  MSFT.US      0.5
        1              Apple Inc  AAPL.US      0.5
        """
        x = pd.DataFrame(
            data={
                "asset name": list(self.names.values()),
                "ticker": list(self.names.keys()),
            }
        )
        x["weights"] = self.weights
        return x

    # Forecasting

    def _test_forecast_period(self, years):
        max_period_years = round(self.period_length / 2)
        if max_period_years < 1:
            raise ValueError(
                f"Time series does not have enough history to forecast. "
                f"Period length is {self.period_length:.2f} years. At least 2 years are required."
            )
        if not isinstance(years, int) or years == 0:
            raise ValueError("years must be an integer number (not equal to zero).")
        if years > max_period_years:
            raise ValueError(
                f"Forecast period {years} years is not credible. "
                f"It should not exceed 1/2 of portfolio history period length {self.period_length / 2} years"
            )

    def percentile_inverse_cagr(
        self,
        distr: str = "norm",
        years: int = 1,
        score: float = 0,
        n: Optional[int] = None,
    ) -> float:
        """
        Compute the percentile rank of a score (CAGR value).

        Percentile rank can be calculated for given distribution type or for hsitorical distribution of CAGR.

        If percentile_inverse of, for example, 0% (CAGR value) is equal to 8% for 1 year time frame
        it means that 8% of the CAGR values in the distribution are negative in 1 year periods. Or in other words
        the probability of getting negative result after 1 year of investments is 8%.

        Parameters
        ----------
        distr: {'norm', 'lognorm', 'hist'}, default 'norm'
            The rate of teturn distribution type.
            'norm' - for normal distribution.
            'lognorm' - for lognormal distribution.
            'hist' - percentiles are taken from the historical data.

        years: int, default 1
            Period length (time frame) in years when CAGR is calculated.

        score: float, default 0
            Score that is compared to the elements in CAGR array.

        n: int, optional
            Number of random time series with the defined distributions (for 'norm' or 'lognorm' only).
            Larger argument values can be used to increase the precision of the calculation. But this will lead
            to slower performance.
            Is not required for historical distribution (dist='hist').
            For 'norm' or 'lognorm' distribution default value n=1000 is used.

        Returns
        -------
        float
            Percentile-position of score (0-100) relative to distribution.

        Examples
        --------
        >>> pf = ok.Portfolio(['SPY.US', 'AGG.US', 'GLD.US'], weights=[.60, .35, .05], rebalancing_period='year')
        >>> pf.percentile_inverse_cagr(distr='lognorm', score=0, years=1, n=5000)
        18.08
        The probability of getting negative result (score=0) in 1 year period for lognormal distribution.
        """
        if distr == "hist":
            cagr_distr = self.get_rolling_cagr(years)
        elif distr in ["norm", "lognorm"]:
            if not n:
                n = 1000
            cagr_distr = self._get_cagr_distribution(
                distr=distr, years=years, n=n
            )
        else:
            raise ValueError('distr should be one of "norm", "lognorm", "hist".')
        return scipy.stats.percentileofscore(cagr_distr, score, kind="rank")

    def percentile_history_cagr(
        self, years: int, percentiles: List[int] = [10, 50, 90]
    ) -> pd.DataFrame:
        """
        Calculate given percentiles for portfolio rolling CAGR distribution from the historical data.

        CAGR - Compound Annual Growth Rate.
        Each percentile is calculated for a period range from 1 year to 'years'.

        Parameters
        ----------
        years: int, default 1
            Max window size for rolling CAGR in the distribution in years.
            It should not exceed 1/2 of the portfolio history period length 'period_length'.

        percentiles: list of int, default [10, 50, 90]
            List of percentiles to be calculated.

        Returns
        -------
        DataFrame
            Table with percentiles values for each period from 1 to 'years'.

        Examples
        --------
        >>> pf = ok.Portfolio(['SPY.US', 'AGG.US', 'GLD.US'], weights=[.60, .35, .05], rebalancing_period='none')
        >>> pf.percentile_history_cagr(years=5, percentiles=[1, 50, 99])
                     1         50        99
        years
        1     -0.231327  0.098693  0.295343
        2     -0.101689  0.091824  0.206471
        3     -0.036771  0.085428  0.157833
        4     -0.007674  0.085178  0.142195
        5      0.030933  0.082865  0.134496
        """
        self._test_forecast_period(years)
        period_range = range(1, years + 1)
        returns_dict = {}
        for percentile in percentiles:
            percentile_returns_list = [
                self.get_rolling_cagr(years * 12).loc[:, self.symbol].quantile(percentile / 100)
                for years in period_range
            ]
            returns_dict.update({percentile: percentile_returns_list})
        df = pd.DataFrame(returns_dict, index=list(period_range))
        df.index.rename("years", inplace=True)
        return df

    def percentile_wealth_history(
        self, years: int = 1, percentiles: List[int] = [10, 50, 90]
    ) -> pd.DataFrame:
        """
        Calculate portfolio wealth index percentiles.

        Percentiles are derived from rolling CAGR historical distribution.
        CAGR - Compound Annual Growth Rate.
        Wealth index (Cumulative Wealth Index) is a time series that presents the value of portfolio over a given
        time period.

        Actual portfolio wealth is adjusted to the last known historical value (from 'wealth_index'). It is useful
        for a chart with historical wealth index and forecasted values.

        Parameters
        ----------
        years: int, default 1
            Time frame for portfolio wealth index percentiles.
            It should not exceed 1/2 of the portfolio history period length 'period_length'.
            Percentiles are calculated for periods from 1 to 'years'.

        percentiles: list of int, default [10, 50, 90]
            List of percentiles to be calculated.

        Returns
        -------
        DataFrame
            Table with portfolio wealth index percentiles for each period from 1 to 'years'.

        Examples
        --------
        >>> pf = ok.Portfolio(['SPY.US', 'AGG.US', 'GLD.US'], weights=[.60, .35, .05], rebalancing_period='month')
        >>> pf.percentile_wealth_history(years=5)
                        10           50           90
        years
        1      3815.660408  4202.758919  4457.210561
        2      3727.946026  4540.888480  5005.291952
        3      3797.214674  4855.631902  5384.216628
        4      4173.503054  5274.584657  6018.571025
        5      4613.287195  5706.343210  6694.576137
        """
        first_value = self.wealth_index[self.symbol].values[-1]
        percentile_returns = self.percentile_history_cagr(
            years=years, percentiles=percentiles
        )
        return first_value * (percentile_returns + 1.0).pow(
            percentile_returns.index.values, axis=0
        )

    def _forecast_preparation(self, years: int):
        self._test_forecast_period(years)
        period_months = years * _MONTHS_PER_YEAR
        # make periods index where the shape is max_period
        start_period = self.last_date.to_period("M")
        end_period = self.last_date.to_period("M") + period_months - 1
        ts_index = pd.period_range(start_period, end_period, freq="M")
        return period_months, ts_index

    def monte_carlo_returns_ts(
        self, distr: str = "norm", years: int = 1, n: int = 100
    ) -> pd.DataFrame:
        """
        Generate portfolio monthly rate of return time series with Monte Carlo simulation.

        Monte Carlo simulation generates n random monthly time series with a given distribution.
        Forecast period should not exceed 1/2 of portfolio history period length.

        First date of forecaseted returns is portfolio last_date.

        Parameters
        ----------
        distr : {'norm', 'lognorm'}, default 'norm'
            Distribution type for rate of return time series.
            'norm' - for normal distribution.
            'lognorm' - for lognormal distribution.

        years : int, default 1
            Forecast period for portfolio monthly rate of return time series.
            It should not exceed 1/2 of the portfolio history period length 'period_length'.

        n : int, default 100
            Number of random rate of return time series to generate with Monte Carlo simulation.

        Returns
        -------
        DataFrame
            Table with n random rate of return monthly time series.

        Examples
        --------
        >>> pf = ok.Portfolio(['SPY.US', 'AGG.US', 'GLD.US'], weights=[.60, .35, .05], rebalancing_period='month')
        >>> pf.monte_carlo_returns_ts(years=8, distr='norm', n=5000)
                     0         1         2     ...      4997      4998      4999
        2021-07 -0.008383 -0.013167 -0.031659  ...  0.046717  0.065675  0.017933
        2021-08  0.038773 -0.023627  0.039208  ... -0.016075  0.034439  0.001856
        2021-09  0.005026 -0.007195 -0.003300  ... -0.041591  0.021173  0.114225
        2021-10 -0.007257  0.003013 -0.004958  ...  0.037057 -0.009689 -0.003242
        2021-11 -0.005006  0.007090  0.020741  ...  0.026509 -0.023554  0.010271
                   ...       ...       ...  ...       ...       ...       ...
        2029-02 -0.065898 -0.003673  0.001198  ...  0.039293  0.015963 -0.050704
        2029-03  0.021215  0.008783 -0.017003  ...  0.035144  0.002169  0.015055
        2029-04  0.002454 -0.016281  0.017004  ...  0.032535  0.027196 -0.029475
        2029-05  0.011206  0.023396 -0.013757  ... -0.044717 -0.025613 -0.002066
        2029-06 -0.016740 -0.007955  0.002862  ... -0.027956 -0.012339  0.048974
        [96 rows x 5000 columns]
        """
        period_months, ts_index = self._forecast_preparation(years)
        # random returns
        if distr == "norm":
            random_returns = np.random.normal(
                self.mean_return_monthly, self.risk_monthly, (period_months, n)
            )
        elif distr == "lognorm":
            std, loc, scale = scipy.stats.lognorm.fit(self.ror)
            random_returns = scipy.stats.lognorm(std, loc=loc, scale=scale).rvs(
                size=[period_months, n]
            )
        else:
            raise ValueError('"distr" must be "norm" (default) or "lognorm".')
        return pd.DataFrame(data=random_returns, index=ts_index)

    def _monte_carlo_wealth(
        self, distr: str = "norm", years: int = 1, n: int = 100
    ) -> pd.DataFrame:
        """
        Generate portfolio wealth index with Monte Carlo simulation.

        Monte Carlo simulation generates n random monthly time series.
        Each wealth index is calculated with rate of return time series of a given distribution.

        Forecast period should not exceed 1/2 of portfolio history period length.
        First date of forecasted returns is portfolio last_date.
        First value for the forecasted wealth indexes is the last historical portfolio index value. It is useful
        for a chart with historical wealth index and forecasted values.

        Parameters
        ----------
        distr : {'norm', 'lognorm'}, default 'norm'
            Distribution type for the rate of return of portfolio.

        years : int, default 1
            Forecast period for portfolio wealth index time series.
            It should not exceed 1/2 of the portfolio history period length 'period_length'.

        n : int, default 100
            Number of random wealth indexes to generate with Monte Carlo simulation.

        Returns
        -------
        DataFrame
            Table with n random wealth indexes monthly time series.

        Examples
        --------
        >>> pf = ok.Portfolio(['SPY.US', 'AGG.US', 'GLD.US'], weights=[.60, .35, .05], rebalancing_period='month')
        >>> pf._monte_carlo_wealth(distr='lognorm', years=5, n=1000)
                         0            1    ...          998          999
        2021-07  3895.377293  3895.377293  ...  3895.377293  3895.377293
        2021-08  3869.854680  4004.814981  ...  3874.455244  3935.913516
        2021-09  3811.125717  3993.783034  ...  3648.925159  3974.103856
        2021-10  4053.024519  4232.141143  ...  3870.099003  4082.189688
        2021-11  4179.544897  4156.839698  ...  3899.249696  4097.003962
        2021-12  4237.030690  4351.305114  ...  3916.639721  4042.011774
        """
        if distr not in ["norm", "lognorm"]:
            raise ValueError('distr should be "norm" (default) or "lognorm".')
        return_ts = self.monte_carlo_returns_ts(distr=distr, years=years, n=n)
        first_value = self.wealth_index[self.symbol].values[-1]
        return Frame.get_wealth_indexes(return_ts, first_value)

    def _get_cagr_distribution(
        self, distr: str = "norm", years: int = 1, n: int = 100,
    ) -> pd.Series:
        """
        Generate CAGR distribution for the rate of return distribution of a given type.

        CAGR is calculated for each of n random returns time series.
        """
        if distr not in ["norm", "lognorm"]:
            raise ValueError('distr should be "norm" (default) or "lognorm".')
        return_ts = self.monte_carlo_returns_ts(distr=distr, years=years, n=n)
        return Frame.get_cagr(return_ts)

    def percentile_distribution_cagr(
        self,
        distr: str = "norm",
        years: int = 1,
        percentiles: List[int] = [10, 50, 90],
        n: int = 10000,
    ) -> Dict[int, float]:
        """
        Calculate percentiles for a given CAGR distribution.

        CAGR - Compound Annual Growth Rate.
        CAGR is calculated for each of n random returns time series of a given distribution. Random time series are
        generated with Monte Carlo simulation.
        CAGR time frame should not exceed 1/2 of portfolio history period length.

        Parameters
        ----------
        distr : {'norm', 'lognorm'}, default 'norm'
            Distribution type for the rate of return of portfolio.
            'norm' - for normal distribution.
            'lognorm' - for lognormal distribution.

        years: int, default 1
            Time frame for portfolio CAGR.
            It should not exceed 1/2 of the portfolio history period length 'period_length'.

        percentiles: list of int, default [10, 50, 90]
            List of percentiles to be calculated.

        n : int, default 10000
            Number of random time series to generate with Monte Carlo simulation.
            Larger argument values can be used to increase the precision of the calculation. But this will lead
            to slower performance.

        Returns
        -------
        dict
            Dictionary {Percentile: value}

        Examples
        --------
        >>> pf = ok.Portfolio(['SPY.US', 'AGG.US', 'GLD.US'], weights=[.60, .35, .05], rebalancing_period='year')
        >>> pf.percentile_distribution_cagr()
        {10: -0.0329600265453808, 50: 0.08247141141668779, 90: 0.21338327078214836}
        Forecast CAGR according to normal distribution within 1 year period.
        >>> pf.percentile_distribution_cagr(years=5)
        {10: 0.030625112922274055, 50: 0.08346815557550402, 90: 0.13902575176654647}
        Forecast CAGR according to normal distribution within 5 year period.
        """
        if distr not in ["norm", "lognorm"]:
            raise ValueError('distr should be "norm" (default) or "lognorm".')
        cagr_distr = self._get_cagr_distribution(
            distr=distr, years=years, n=n
        )
        results = {}
        for percentile in percentiles:
            value = cagr_distr.quantile(percentile / 100)
            results.update({percentile: value})
        return results

    def percentile_wealth(
        self,
        distr: str = "norm",
        years: int = 1,
        percentiles: List[int] = [10, 50, 90],
        today_value: Optional[int] = None,
        n: int = 1000,
    ) -> Dict[int, float]:
        """
        Calculate percentiles for portfolio wealth indexes distribution.

        Portfolio wealth indexes are derived from the rate of return time series of a given distribution type.

        Parameters
        ----------
        distr : {'hist', 'norm', 'lognorm'}, default 'norm'
            Distribution type for the rate of return of portfolio.
            'norm' - for normal distribution.
            'lognorm' - for lognormal distribution.
            'hist' - percentiles are taken from the historical data.

        years : int, default 1
            Investment period length to calculate wealth index.
            It should not exceed 1/2 of the portfolio history period length 'period_length'.

        percentiles : list of int, default [10, 50, 90]
            List of percentiles to be calculated.

        today_value :  int, optional
            Initial value of the wealth index.
            If today_value is None the last value of the historical wealth indexes is taken. It can be useful to plot
            the forecast of wealth index togeather with the hitorical data.

        n : int, default 1000
            Number of random time series to generate with Monte Carlo simulation (for 'norm' or 'lognorm' only).
            Larger argument values can be used to increase the precision of the calculation. But this will lead
            to slower performance.
            Is not required for historical distribution (dist='hist').

        Returns
        -------
        dict
            Dictionary {Percentile: value}

        Examples
        --------
        >>> pf = ok.Portfolio(['SPY.US', 'AGG.US', 'GLD.US'], weights=[.60, .35, .05], rebalancing_period='year')
        >>> pf.percentile_wealth(distr='hist', years=5, today_value=1000, n=5000)
        {10: 1228.3741255659957, 50: 1491.7857161011104, 90: 1745.1130920663286}
        Percentiles values for the wealth index 5 years forecast if the initial value is 1000.
        """
        if distr == "hist":
            results = (
                self.percentile_wealth_history(years=years, percentiles=percentiles)
                .iloc[-1]
                .to_dict()
            )
        elif distr in ["norm", "lognorm"]:
            results = {}
            wealth_indexes = self._monte_carlo_wealth(
                distr=distr, years=years, n=n
            )
            for percentile in percentiles:
                value = wealth_indexes.iloc[-1, :].quantile(percentile / 100)
                results.update({percentile: value})
        else:
            raise ValueError('distr should be "norm", "lognorm" or "hist".')
        if today_value:
            modifier = today_value / self.wealth_index[self.symbol].values[-1]
            results.update((x, y * modifier) for x, y in results.items())
        return results

    # distributions
    @property
    def skewness(self) -> pd.Series:
        """
        Compute expanding skewness time series for portfolio rate of return.

        For normally distributed data, the skewness should be about zero.
        A skewness value greater than zero means that there is more weight in the right tail of the distribution.

        Returns
        -------
        Series
            Expanding skewness time series

        Examples
        --------
        >>> pf = ok.Portfolio(['BND.US'])
        >>> pf.skewness
        Date
        2008-05   -0.134193
        2008-06   -0.022349
        2008-07    0.081412
        2008-08   -0.020978
                     ...
        2021-04    0.441430
        2021-05    0.445772
        2021-06    0.437383
        2021-07    0.425247
        Freq: M, Name: portfolio_8378.PF, Length: 159, dtype: float64

        >>> import matplotlib.pyplot as plt
        >>> pf.skewness.plot()
        >>> plt.show()
        """
        return Frame.skewness(self.ror)

    def skewness_rolling(self, window: int = 60):
        """
        Compute rolling skewness of the return time series.

        For normally distributed rate of return, the skewness should be about zero.
        A skewness value greater than zero means that there is more weight in the right tail of the distribution.

        Parameters
        ----------
        window : int, default 60
            Size of the moving window in months.
            The window size should be at least 12 months.

        Returns
        -------
        Series
            Expanding skewness time series

        Examples
        --------
        >>> pf = ok.Portfolio(['BND.US'])
        >>> pf.skewness_rolling(window=12*10)
        Date
        2017-04    0.464916
        2017-05    0.446095
        2017-06    0.441211
        2017-07    0.453947
        2017-08    0.464805
        ...
        2021-02    0.007622
        2021-03    0.000775
        2021-04    0.002308
        2021-05    0.022543
        2021-06   -0.006534
        2021-07   -0.012192
        Freq: M, Name: portfolio_8378.PF, dtype: float64

        >>> import matplotlib.pyplot as plt
        >>> pf.skewness_rolling(window=12*10).plot()
        >>> plt.show()
        """
        return Frame.skewness_rolling(self.ror, window=window)

    @property
    def kurtosis(self):
        """
        Calculate expanding Fisher (normalized) kurtosis time series for portfolio rate of return.

        Kurtosis is a measure of whether the rate of return are heavy-tailed or light-tailed
        relative to a normal distribution.
        It should be close to zero for normally distributed rate of return.
        Kurtosis is the fourth central moment divided by the square of the variance.

        Returns
        -------
        Series
            Expanding kurtosis time series

        Examples
        --------
        >>> pf = ok.Portfolio(['BND.US'])
        >>> pf.kurtosis
        Date
        2008-05   -0.815206
        2008-06   -0.718330
        2008-07   -0.610741
        2008-08   -0.534105
                     ...
        2021-04    2.821322
        2021-05    2.855267
        2021-06    2.864717
        2021-07    2.850407
        Freq: M, Name: portfolio_4411.PF, Length: 159, dtype: float64

        >>> import matplotlib.pyplot as plt
        >>> pf.kurtosis.plot()
        >>> plt.show()
        """
        return Frame.kurtosis(self.ror)

    def kurtosis_rolling(self, window: int = 60):
        """
        Calculate rolling Fisher (normalized) kurtosis time series for portfolio rate of return.

        Kurtosis is a measure of whether the rate of return are heavy-tailed or light-tailed
        relative to a normal distribution.
        It should be close to zero for normally distributed rate of return.
        Kurtosis is the fourth central moment divided by the square of the variance.

        Parameters
        ----------
        window : int, default 60
            Size of the moving window in months.
            The window size should be at least 12 months.

        Returns
        -------
        Series
            Expanding kurtosis time series.

        Examples
        --------
        >>> pf = ok.Portfolio(['BND.US'])
        >>> pf.kurtosis_rolling(window=12*10)
        Date
        2017-04    4.041599
        2017-05    4.133518
        2017-06    4.165099
        2017-07    4.205125
        2017-08    4.313773
        ...
        2021-03    0.362184
        2021-04    0.409680
        2021-05    0.455760
        2021-06    0.457315
        2021-07    0.496168
        Freq: M, Name: portfolio_4411.PF, dtype: float64

        >>> import matplotlib.pyplot as plt
        >>> pf.kurtosis_rolling(window=12*10).plot()
        >>> plt.show()
        """
        return Frame.kurtosis_rolling(self.ror, window=window)

    @property
    def jarque_bera(self) -> Dict[str, float]:
        """
        Perform Jarque-Bera test for normality of portfolio returns time series.

        Jarque-Bera shows whether the returns have the skewness and kurtosis
        matching a normal distribution (null hypothesis or H0).

        Returns
        -------
        dict
            Jarque-Bera test statistics and p-value.

        Notes
        -----
        Test returns statistics (first row) and p-value (second row).
        p-value is the probability of obtaining test results, under the assumption that the null hypothesis is correct.
        In general, a large Jarque-Bera statistics and tiny p-value indicate that null hypothesis is rejected
        and the time series are not normally distributed.

        Examples
        --------
        >>> pf = ok.Portfolio(['BND.US'])
        >>> pf.jarque_bera
        {'statistic': 58.27670538027455, 'p-value': 2.2148949341271873e-13}
        """
        return Frame.jarque_bera_series(self.ror)

    def kstest(self, distr: str = "norm") -> Dict[str, float]:
        """
        Perform one sample Kolmogorov-Smirnov test on portfolio returns and evaluate goodness of fit
        for a given distribution.

        The one-sample Kolmogorov-Smirnov test compares the rate of return time series against a given distribution.

        Returns
        -------
        dict
            Kolmogorov-Smirnov test statistics and p-value.

        Parameters
        ----------
        distr : {'norm', 'lognorm'}, default 'norm'
            The name of a distribution to fit.
            'norm' - for normal distribution.
            'lognorm' - for lognormal distribution.


        Notes
        -----
        Like in Jarque-Bera test returns statistic (first row) and p-value (second row).
        Null hypotesis (two distributions are similar) is not rejected when p-value is high enough.
        5% threshold can be used.

        Examples
        --------
        >>> pf = ok.Portfolio(['GLD.US'])
        >>> pf.kstest(distr='lognorm')
        {'statistic': 0.05001344986084533, 'p-value': 0.6799422889377373}

        >>> pf.kstest(distr='norm')
        {'statistic': 0.09528000069992831, 'p-value': 0.047761781235967415}

        Kolmogorov-Smirnov test shows that GLD rate of return time series fits lognormal distribution
        better than normal one.
        """
        return Frame.kstest_series(self.ror, distr=distr)

    def get_sharpe_ratio(self, rf_return: float = 0) -> float:
        """
        Calculate Sharpe ratio.

        The Sharpe ratio is the average annual return in excess of the risk-free rate
        per unit of risk (annualized standard deviation).

        Risk-free rate should be taken according to the Portfolio base currency.

        Parameters
        ----------
        rf_return : float, default 0
            Risk-free rate of return.

        Returns
        -------
        float

        Examples
        --------
        >>> pf = ok.Portfolio(['VOO.US', 'BND.US'], weights=[0.40, 0.60])
        >>> pf.get_sharpe_ratio(rf_return=0.04)
        0.7412193684695373
        """
        return ratios.get_sharpe_ratio(
            pf_return=self.mean_return_annual,
            rf_return=rf_return,
            std_deviation=self.risk_annual)

    def plot_percentiles_fit(
        self, distr: str = "norm", figsize: Optional[tuple] = None
    ) -> None:
        """
        Generate a quantile-quantile (Q-Q) plot of portfolio monthly rate of return against quantiles of a given
        theoretical distribution.

        A q-q plot is a plot of the quantiles of the portfolio rate of return historical data
        against the quantiles of a given theoretical distribution.

        Parameters
        ----------
        distr : {'norm', 'lognorm'}, default 'norm'
            The name of a distribution to fit.
            'norm' - for normal distribution.
            'lognorm' - for lognormal distribution.

        figsize : (float, float), optional
            Width and height of plot in inches.
            If None default matplotlib figsize value is used.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(['SPY.US', 'AGG.US', 'GLD.US'], weights=[.60, .35, .05], rebalancing_period='year')
        >>> pf.plot_percentiles_fit(distr='lognorm')
        >>> plt.show()
        """
        plt.figure(figsize=figsize)
        if distr == "norm":
            scipy.stats.probplot(self.ror, dist=distr, plot=plt)
        elif distr == "lognorm":
            scipy.stats.probplot(
                self.ror,
                sparams=(scipy.stats.lognorm.fit(self.ror)),
                dist=distr,
                plot=plt,
            )
        else:
            raise ValueError('distr should be "norm" (default) or "lognorm".')
        plt.show()

    def plot_hist_fit(self, distr: str = "norm", bins: int = None) -> None:
        """
        Plot historical distribution histogram for ptrtfolio monthly rate of return time series
        and theoretical PDF (Probability Distribution Function).

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(['SP500TR.INDX'])
        >>> pf.plot_hist_fit(distr='norm')
        >>> plt.show()
        """
        data = self.ror
        # Plot the histogram
        plt.hist(data, bins=bins, density=True, alpha=0.6, color="g")
        # Plot the PDF.Probability Density Function
        xmin, xmax = plt.xlim()
        x = np.linspace(xmin, xmax, 100)
        if distr == "norm":  # Generate PDF
            mu, std = scipy.stats.norm.fit(data)
            p = scipy.stats.norm.pdf(x, mu, std)
        elif distr == "lognorm":
            std, loc, scale = scipy.stats.lognorm.fit(data)
            mu = np.log(scale)
            p = scipy.stats.lognorm.pdf(x, std, loc, scale)
        else:
            raise ValueError('distr must be "norm" (default) or "lognorm".')
        plt.plot(x, p, "k", linewidth=2)
        title = "Fit results: mu = %.3f,  std = %.3f" % (mu, std)
        plt.title(title)
        plt.show()

    def plot_forecast(
        self,
        distr: str = "norm",
        years: int = 5,
        percentiles: List[int] = [10, 50, 90],
        today_value: Optional[int] = None,
        n: int = 1000,
        figsize: Optional[tuple] = None,
    ) -> plt.axes:
        """
        Plot forecasted ranges of wealth indexes (lines) for a given set of percentiles.
        Historical wealth index is shown in the same chart.

        Parameters
        ----------
        distr : {'hist', 'norm', 'lognorm'}, default 'norm'
            Distribution type for the rate of return of portfolio.
            'norm' - for normal distribution.
            'lognorm' - for lognormal distribution.
            'hist' - percentiles are taken from the historical data.

        years : int, default 1
            Investment period length to calculate wealth index.
            It should not exceed 1/2 of the portfolio history period length 'period_length'.

        percentiles : list of int, default [10, 50, 90]
            List of percentiles to be calculated.

        today_value :  int, optional
            Initial value of the wealth index.
            If today_value is None the last value of the historical wealth indexes is taken. It can be useful to plot
            the forecast of wealth index togeather with the hitorical data.

        n : int, default 1000
            Number of random time series to generate with Monte Carlo simulation (for 'norm' or 'lognorm' only).
            Larger argument values can be used to increase the precision of the calculation. But this will lead
            to slower performance.
            Is not required for historical distribution (dist='hist').

        Returns
        -------
        Axes : 'matplotlib.axes._subplots.AxesSubplot'

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(['SPY.US', 'AGG.US', 'GLD.US'], weights=[.60, .35, .05], rebalancing_period='year')
        >>> pf.plot_forecast()
        >>> plt.show()
        """
        wealth = self.wealth_index
        x1 = self.last_date
        x2 = x1.replace(year=x1.year + years)
        y_start_value = wealth[self.symbol].iloc[-1]
        y_end_values = self.percentile_wealth(
            distr=distr, years=years, percentiles=percentiles, n=n
        )
        if today_value:
            modifier = today_value / y_start_value
            wealth *= modifier
            y_start_value = y_start_value * modifier
            y_end_values.update((x, y * modifier) for x, y in y_end_values.items())
        fig, ax = plt.subplots(figsize=figsize)
        ax.plot(
            wealth.index.to_timestamp(),
            wealth[self.symbol],
            linewidth=1,
            label="Historical data",
        )
        for percentile in percentiles:
            x, y = [x1, x2], [y_start_value, y_end_values[percentile]]
            if percentile == 50:
                ax.plot(x, y, color="blue", linestyle="-", linewidth=2, label="Median")
            else:
                ax.plot(
                    x,
                    y,
                    linestyle="dashed",
                    linewidth=1,
                    label=f"Percentile {percentile}",
                )
        ax.legend(loc="upper left")
        return ax

    def plot_forecast_monte_carlo(
        self,
        distr: str = "norm",
        years: int = 1,
        n: int = 20,
        figsize: Optional[tuple] = None,
    ) -> None:
        """
        Plot Monte Carlo simulation for portfolio wealth indexes together with historical wealth index.

        Random wealth indexes are generated according to a given distribution.

        Parameters
        ----------
        distr : {'norm', 'lognorm'}, default 'norm'
            Distribution type for the rate of return of portfolio.
            'norm' - for normal distribution.
            'lognorm' - for lognormal distribution.

        years : int, default 1
            Investment period length for new wealth indexes
            It should not exceed 1/2 of the portfolio history period length 'period_length'.

        n : int, default 20
            Number of random wealth indexes to generate with Monte Carlo simulation.

        figsize : (float, float), optional
            Width, height in inches.
            If None default matplotlib figsize value is used.

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> pf = ok.Portfolio(['SPY.US', 'AGG.US', 'GLD.US'], weights=[.60, .35, .05], rebalancing_period='year')
        >>> pf.plot_forecast_monte_carlo(years=5, distr='lognorm', n=100)
        >>> plt.show()
        """
        s1 = self.wealth_index
        s2 = self._monte_carlo_wealth(distr=distr, years=years, n=n)
        s1[self.symbol].plot(legend=None, figsize=figsize)
        for n in s2:
            s2[n].plot(legend=None)



from pytest import mark
from pytest import approx
import pandas as pd


@mark.asset
@mark.smoke
def test_get_symbol_data(init_asset_spy):
    assert init_asset_spy.name == "SPDR S&P 500 ETF Trust"
    assert init_asset_spy.country == "USA"
    assert init_asset_spy.currency == "USD"
    assert init_asset_spy.type == "ETF"
    assert init_asset_spy.inflation == "USD.INFL"
    assert init_asset_spy.first_date == pd.to_datetime("1993-02")


def test_close_daily(init_asset_spy):
    assert init_asset_spy.close_daily.loc['2000-01-20'] == 144.75


def test_close_monthly(init_asset_spy):
    assert init_asset_spy.close_monthly.loc['2000-01'] == 139.5625


def test_adj_close(init_asset_spy):
    assert init_asset_spy.adj_close.loc['2000-01-20'] == approx(97.0629, rel=1e-2)


def test_price(init_asset_spy):
    assert type(init_asset_spy.price) == float


def test_dividends(init_asset_spy):
    assert init_asset_spy.dividends["2019"].iloc[-1] == 1.57


def test_nav_ts(init_asset_pif):
    assert init_asset_pif.nav_ts[0] == 101820352.18



from typing import Dict

from io import StringIO
import json

import pandas as pd
import numpy as np

from .api_methods import API
from .namespaces import no_dividends_namespaces


class QueryData:
    """
    Set of methods to get symbols data from API.
    """

    @staticmethod
    def get_symbol_info(symbol: str) -> Dict[str, str]:
        json_input = API.get_symbol_info(symbol)
        return json.loads(json_input)

    @staticmethod
    def csv_to_series(csv_input: str, period: str) -> pd.Series:
        ts = pd.read_csv(
            StringIO(csv_input),
            delimiter=",",
            index_col=0,
            parse_dates=[0],
            dtype={1: np.float64},
            engine="python",
        )
        if not ts.empty:
            ts.index = ts.index.to_period(period.upper())
            ts = ts.squeeze("columns")
        return ts

    @staticmethod
    def get_macro_ts(
        symbol: str, first_date: str = "1913-01-01", last_date: str = "2100-01-01"
    ) -> pd.Series:
        """
        Requests API for Macroeconomic indicators time series (monthly data).
        - Inflation time series
        - Bank rates time series
        """
        csv_input = API.get_macro(
            symbol=symbol, first_date=first_date, last_date=last_date
        )
        return QueryData.csv_to_series(csv_input, period="M")

    @staticmethod
    def get_ror(
        symbol: str,
        first_date: str = "1913-01-01",
        last_date: str = "2100-01-01",
        period="M",
    ) -> pd.Series:
        """
        Requests API for rate of return time series.
        """
        csv_input = API.get_ror(
            symbol=symbol, first_date=first_date, last_date=last_date, period=period
        )
        return QueryData.csv_to_series(csv_input, period)

    @staticmethod
    def get_nav(
        symbol: str,
        first_date: str = "1913-01-01",
        last_date: str = "2100-01-01",
        period="M",
    ) -> pd.Series:
        """
        NAV time series for funds (works for PIF namespace only).
        """
        csv_input = API.get_nav(
            symbol=symbol, first_date=first_date, last_date=last_date, period=period
        )
        return QueryData.csv_to_series(csv_input, period=period)

    @staticmethod
    def get_close(
        symbol: str,
        first_date: str = "1913-01-01",
        last_date: str = "2100-01-01",
        period="M",
    ) -> pd.Series:
        """
        Gets 'close' time series for a ticker.
        """
        csv_input = API.get_close(
            symbol=symbol, first_date=first_date, last_date=last_date, period=period
        )
        return QueryData.csv_to_series(csv_input, period)

    @staticmethod
    def get_adj_close(
        symbol: str,
        first_date: str = "1913-01-01",
        last_date: str = "2100-01-01",
        period="M",
    ) -> pd.Series:
        """
        Gets 'adjusted close' time series for a ticker.
        """
        csv_input = API.get_adjusted_close(
            symbol=symbol, first_date=first_date, last_date=last_date, period=period
        )
        return QueryData.csv_to_series(csv_input, period)

    @staticmethod
    def get_dividends(
        symbol: str, first_date: str = "1913-01-01", last_date: str = "2100-01-01",
    ) -> pd.Series:
        """
        Dividends time series daily data (dividend payment day should be considered).
        """
        if symbol.split(".", 1)[-1] not in no_dividends_namespaces():
            csv_input = API.get_dividends(
                symbol, first_date=first_date, last_date=last_date
            )
            ts = QueryData.csv_to_series(csv_input, period="D")
        else:
            # make empty time series when no dividends
            ts = pd.Series(dtype=float)
            ts.rename(symbol, inplace=True)
        return ts

    @staticmethod
    def get_live_price(symbol: str) -> float:
        price = API.get_live_price(symbol)
        return float(price)



from collections import namedtuple

default_ticker = "SPY.US"
default_namespace = "US"
default_tickers_list = ["SPY.US", "BND.US"]  # required in single_period.py
default_macro = "RUB.INFL"

_MONTHS_PER_YEAR = 12
PeriodLength = namedtuple("PeriodLength", "years months")
